# Python compatibility:
from __future__ import absolute_import

from importlib_metadata import PackageNotFoundError, version

# Setup tools:
from packaging.version import parse as parse_version

try:
    _vps_v = version('visaplan.plone.search')
except PackageNotFoundError:
    getExcludeFromNav = 'exclude_from_nav'  # visaplan.plone.search v1.8.0+
    getExcludeFromSearch = 'getExcludeFromSearch'  # WIP: change index name
else:
    _vps_v = parse_version(_vps_v)
    if _vps_v < parse_version('1.8.0.dev1'):
        raise ImportError('Found visaplan.plone.search %s; '
                 'we need 1.8+, if any!' % (_vps_v,))
    # visaplan:
    from visaplan.plone.search.idxnames import (
        getExcludeFromNav,
        getExcludeFromSearch,
        )



# -*- coding: utf-8 -*-
# Python compatibility:
from __future__ import absolute_import

from importlib_metadata import PackageNotFoundError
from importlib_metadata import version as pkg_version

# Standard library:
from functools import wraps

try:
    pkg_version('simplejson')
except PackageNotFoundError:
    # Standard library:
    from json import dumps
else:
    # 3rd party:
    from simplejson import dumps


def returns_json(raw):
    """
    Decorate the given function to ...
    - convert the result to JSON,
      and
    - add the appropriate HTTP headers

    This is for JSON data requested by a client ("TTW");
    DON'T apply it to functions which are used in a view to build other contents
    around it, as the HTTP headers will be incorrect;
    the user agent might try to parse your HTML page as JSON and fail.

    For this purpose, see the @headless_json decorator (below).
    """
    @wraps(raw)
    def wrapped(self, **kwargs):
        dic = raw(self, **kwargs)
        txt = dumps(dic)
        context = self.context
        setHeader = context.REQUEST.RESPONSE.setHeader
        setHeader('Content-Type', 'application/json; charset=utf-8')
        setHeader('Content-Length', len(txt))
        return txt
    return wrapped


def headless_json(raw):
    """
    Decorate the given function to convert the result to JSON. Period.

    This is for JSON data used internally;
    for public views which serve data to Javascript code you'll prefer
    the @returns_json decorator (above).

    >>> @headless_json
    ... def f(): return dict(eins=2, drei='vier')
    >>> f()
    '{"drei": "vier", "eins": 2}'
    """
    @wraps(raw)
    def wrapped(*args, **kwargs):
        dic = raw(*args, **kwargs)
        txt = dumps(dic)
        return txt
    return wrapped


if __name__ == '__main__':
   import doctest
   doctest.testmod()



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
# Python compatibility:
from __future__ import absolute_import

from importlib_metadata import PackageNotFoundError, version

try:
    version('Products.LinguaPlone')
except PackageNotFoundError:
    class AlreadyTranslated(Exception):
        """
        Dummy exception; Products.LinguaPlone is not installed
        """
else:
    # Zope:
    from Products.LinguaPlone.I18NBaseObject import AlreadyTranslated


class CantAddTranslationReference(AlreadyTranslated):
    """
    LinguaPlone has refused to add a translation reference,
    and we'd like to provide some more useful information
    """
    def __init__(self, o, canonical):
        AlreadyTranslated.__init__(self, o)
        try:
            o_lang = o.Language()
        except Exception as e:
            o_lang = e
        try:
            canonical_lang = canonical.Language()
        except Exception as e:
            canonical_lang = e
        self._raw_mapping = {
            'o': o,
            'o_lang': o_lang,
            'canonical': canonical,
            'canonical_lang': canonical_lang,
            'same_lang': o_lang == canonical_lang,
            'same_o': canonical is o,
            }

    def __str__(self):
        return ("CantAddTranslationReference(%(o)r,\n"
              """                            %(canonical)r):
        1st lang = %(o_lang)r, 2nd lang = %(canonical_lang)r
        (same language: %(same_lang)r; same object: %(same_o)r)
        """).strip() % self._raw_mapping

    def __repr__(self):
        return ("<CantAddTranslationReference(%(o)r, %(canonical)r)>"
                % self._raw_mapping)



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
Kontext-Helferlein:

Funktionen, die im Rahmen eines Aufruf-Kontexts hilfreich sind
(also nicht über den aktuellen Request hinaus)
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types

# Standard library:
from traceback import extract_stack

# Zope:
from Products.CMFCore.utils import getToolByName

# Plone:
from Products.CMFPlone import PloneMessageFactory as pmf

# visaplan:
from visaplan.tools.coding import safe_decode
from visaplan.tools.minifuncs import check_kwargs, gimme_False

# Local imports:
from visaplan.plone.tools._have import HAS_ZOPE_I18N

# Logging / Debugging:
import logging
from visaplan.plone.tools.log import getLogSupport
from visaplan.tools.debug import asciibox, log_or_trace

if HAS_ZOPE_I18N:
    # Zope:
    from zope.i18n import translate as z3translate

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"
VERSION = (0,
           5,   # getMessenger; make_permissionChecker improved
           )

__all__ = ['getActiveLanguage',
           'getActiveLanguage_unchecked',
           'getSupportedLanguageTuples',
           'make_translator',
           'make_pathByUIDGetter',
           'make_permissionChecker',
           'make_timeFormatter',
           'make_userdetector',
           'make_SessionDataProxy',
           'message',       # uses safe_decode
           'getMessenger',  # accepts a decoder argument
           'getbrain',
           'make_brainGetter',
           'parents',
           'parent_brains',
           'get_parent',
           'getPath',
           'get_published_templateid',
           ]


def getActiveLanguage(context):
    """
    Gib den Code der aktiven Sprache zurück;
    wie der Adapter langcode (aus tomcom.adapter), aber "sicherer"
    """
    language_tool = getToolByName(context, 'portal_languages')
    codes = language_tool.getSupportedLanguages()
    request = context.REQUEST
    cookies = request.cookies
    if cookies:
        la = cookies.get('I18N_LANGUAGE')
        if la and la in codes:
            return la
    la = request.get('LANGUAGE')
    if la and la in codes:
        return la
    return language_tool.getDefaultLanguage()


def getActiveLanguage_unchecked(context):
    """
    Return the code of the currently active language;
    like getActiveLanguage (above), but without comparison to supported
    languages, for speed reasons (e.g. for use in cache key functions)
    """
    request = context.REQUEST
    cookies = request.cookies
    if cookies:
        la = cookies.get('I18N_LANGUAGE')
        if la:
            return la
    la = request.get('LANGUAGE')
    if not la:
        pl = getToolByName(context, 'portal_languages')
        la = pl.getDefaultLanguage()
    return la


def getSupportedLanguageTuples(context):
    """
    Return a list of language tuples, for the languages supported in this site
    """
    language_tool = getToolByName(context, 'portal_languages')
    getName = language_tool.getNameForLanguageCode
    liz = []
    for langCode in language_tool.getSupportedLanguages():
        liz.append((langCode, getName(langCode)))
    return liz


def make_translator(context, domain=None, target_language=None):
    """
    Create a function which calls zope.i18n.translate with sensible defaults
    """
    if not HAS_ZOPE_I18N:
        def dummy(msgid,
                  domain=domain,
                  mapping=None,
                  target_language=target_language,
                  default=None):
            """
            translation dummy; zope.i18n not installed
            """
            return default or msgid
        return dummy

    if not domain:
        domain = 'plone'
    if not target_language:
        target_language = getActiveLanguage(context)
    def translate(msgid,
                  domain=domain,
                  mapping=None,
                  target_language=target_language,
                  default=None):
        if not domain:
            domain = 'plone'
        if not target_language:
            target_language = getActiveLanguage(context)
        return z3translate(msgid, domain, mapping,
                           context, target_language, default)
    return translate


def make_permissionChecker(context=None, getAdapter=None,
                           checkperm=None, verbose=None):
    """
    Create a function to check the given permission.

    Options:

    context -- usually the only one given.
               If so, the usual membership_tool.checkPermission method is used,
               with a context default value injected.

    Please specify all following options by name:

    getAdapter -- deprecated;
                  if given, it is used to get a 'checkperm' adapter
                  (which is not present in standard Plone and uses a fixed
                  context argument, which is the one bound to that getAdapter
                  method)
    checkperm -- rarely needed: You might want to use a special permission
                 checking function and add the logging facility.
    verbose -- if True, every check for a permission will be logged.

    Usage:

      checkperm = make_permissionChecker(context)
      if not checkperm('Manage portal'):
          raise Unauthorized
    """
    if checkperm is None:
        if getAdapter is not None:
            try:  # traditional / deprecated usage:
                checkperm = getAdapter('checkperm')
            except:
                pass
    if checkperm is None:
        if context is None:
            raise ValueError('At least the context is needed!')
        membership_tool = getToolByName(context, 'portal_membership')
        _cp = membership_tool.checkPermission
        def checkperm(permission, context=context):
            return _cp(permission, context)

    if verbose:
        # bisher nur mit Modul-, nicht mit Funktions-/Methodeninformation:
        logger = getLogSupport(stack_limit=4)[0]
        log = logger.info
        def cp_verbose(permission):
            val = checkperm(permission)
            log('checkperm(%(permission)r) --> %(val)r', locals())
            return val
        return cp_verbose
    return checkperm


def make_timeFormatter(context, longFormat=False):
    """
    Wie der Adapter totime
    """
    util = getToolByName(context, 'translation_service')
    func = util.ulocalized_time
    request = context.REQUEST

    def format(time, longFormat=longFormat):
        return func(time, longFormat,
                    context=context,
                    domain='plonelocales',
                    request=request)
    return format


def make_SessionDataProxy(context):
    """
    Stellt die Sitzung als Dict. zur Verfügung;
    für fehlende Schlüssel wird None zurückgegeben.
    """

    session = context.REQUEST.SESSION
    # session is incomplete, in lacking a __contains__ method;
    # by this trick we prevent python-modernize from breaking our code:
    has_key = session.has_key

    class SDProxy(dict):
        """
        session: in Closure
        """
        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                if has_key(key):  # noqa; see above
                    val = session.get(key)
                    dict.__setitem__(self, key, val)
                else:
                    val = None
                    session.set(key, val)
                return val

        def __setitem__(self, key, val):
            dict.__setitem__(self, key, val)
            session.set(key, val)

        def __delitem__(self, key):
            try:
                del session[key]
            except KeyError:
                pass
            try:
                del self[key]
            except KeyError:
                pass

    return SDProxy()


class SessionDataProxy(dict):
    """
    Stellt die Sitzung als Dict. zur Verfügung;
    für fehlende Schlüssel wird None zurückgegeben.
    """
    def __init__(self, context):
        # sdm = getToolByName(context, 'session_data_manager')
        # self.session = sdm.getSessionData(create=True)
        self.session = context.REQUEST.SESSION

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            val = self.session.get(key)
            dict.__setitem__(self, key, val)
            return val

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)
        self.session[key] = val

    def __delitem__(self, key):
        try:
            del self.session[key]
        except KeyError:
            pass
        try:
            del self[key]
        except KeyError:
            pass


def make_userdetector(ids, splitter=None, getid=True, verbose=None):
    """
    Gib eine Funktion zurück, die True zurückgibt, wenn ein Benutzer mit einer
    der angegebenen IDs angemeldet ist, und ansonsten False.
    Wenn die Liste leer ist, wird eine Funktion zurückgegeben, die ohne
    weiteres Federlesens False zurückgibt.

    ids -- die Benutzer-IDs
    splitter -- wenn ids ein String ist, übergeben an die split-Methode;
                None (Vorgabe): jeder Leerraum
    getid -- wenn True (Vorgabe), wird die Benutzer-ID oder None zurückgegeben,
             sonst True oder False
    verbose -- wenn True, wird beim Erzeugen des Detektors eine Zeile nach
               StdOut geschrieben. Vorgabe ist True, wenn <splitter> einen
               möglicherweise sinnlosen Wert hat, sonst False.

    Verwendung:

    >>> watched_user = make_userdetector('heinz willy')
    >>> if watched_user(context=context):
    >>>     # angemeldeter Benutzer ist 'heinz' oder 'willy'
    >>>     ...
    """
    if not ids:
        return gimme_False
    elif isinstance(ids, six_string_types):
        ids = ids.split(splitter)
        if not ids:
            return gimme_False
    IDS = frozenset(ids)
    if verbose is None:
        verbose = (splitter is not None
                   and len(splitter) > 1
                   )
    if verbose:
        print('*' * 79)
        print('make_userdetector(%s)' % arginfo(sorted(IDS), getid=getid))
        print('*' * 79)

    def get_userid(**kwargs):
        x = kwargs.get('member_id', None)
        if x:
            return x
        x = kwargs.get('member', None)
        if x:
            return x.getId()
        auth = kwargs.get('auth')
        if auth is None:
            getAdapter = kwargs.get('getAdapter')
            if getAdapter is not None:  # olde / depretated usage:
                try:
                    auth = getAdapter('auth')
                except:
                    pass
            if auth is None:
                context = kwargs.get('context')
                if context is None:
                    raise ValueError('Please specify the context!')
                membership_tool = getToolByName(context, 'portal_membership')
                auth = membership_tool.getAuthenticatedMember
        member = auth()
        return member.getId()

    def get_watched_userid(**kwargs):
        usid = get_userid(**kwargs)
        if usid and usid in IDS:
            return usid
        return None

    def user_is_one_of(**kwargs):
        usid = get_userid(**kwargs)
        if usid and usid in IDS:
            return True
        return False

    if getid:
        return get_watched_userid
    else:
        return user_is_one_of


def decorated_tool(context, toolname, limit_get_delta=0, limit=3):
    """
    Gib eine "dekorierte" Fassung des übergebenen Tools zurück

    Argumente:

    -- context, toolname -- weitergereicht an getToolByName
    -- limit_get_delta -- eine Ganzzahl, die das limit-Argument für
                    extract_stack für die *Beschaffung* des Tools gegenüber
                    seiner *Verwendung* erhöht.
                    Bei Aufruf durch einen Adapter ist 1 zu übergeben.
    -- limit -- weitergereicht an extract_stack
    """
    logger = logging.getLogger('decorated_tool %(toolname)r' % locals())
    def caller_info(limit=limit):
        raw_info = extract_stack(limit=limit)
        filename, lineno, funcname = raw_info[0][:3]
        if filename.endswith('.pyc'):
            filename = filename[:-1]
        if funcname:
            prefix = '%(filename)s[%(funcname)s]: %(lineno)d'
        else:
            prefix = '%(filename)s: %(lineno)d'
        return prefix % locals()

    ci = caller_info(limit+limit_get_delta)
    logger.info('GET:  ' + ci)
    print('*** ' + ci)
    print('*** getting tool %(toolname)r' % locals())

    tool = getToolByName(context, toolname)
    def decorated(*args, **kwargs):
        ci = caller_info()
        logger.info('CALL: ' + ci)
        print('***' + ci + ':')
        print(asciibox((toolname+'(',) + args, kwargs=kwargs))
        res = tool(*args, **kwargs)
        print('...' + ci + '.')
        return res

    decorated.__doc__ = '%(toolname)s tool (decorated)' % locals()
    return decorated


def message(context, message, messageType='info', mapping=None):
    """
    Ersetzt den gleichnamigen Tomcom-Adapter
    """
    pu = getToolByName(context, 'plone_utils')
    pu.addPortalMessage(pmf(safe_decode(message),
                            mapping=mapping),
                        messageType)


def getMessenger(context, decode=safe_decode):
    """
    Return a 'message' function which doesn't require
    (nor accept) a context argument
    """
    pu = getToolByName(context, 'plone_utils')
    apm = pu.addPortalMessage
    def message(message, messageType='info', mapping=None):
        """
        Ersetzt den gleichnamigen Tomcom-Adapter
        """
        apm(pmf(decode(message),
                mapping=mapping),
            messageType)
    return message


def getbrain(context, uid):
    """
    Ersetzt den gleichnamigen Tomcom-Adapter
    """
    pc = getToolByName(context, 'portal_catalog')._catalog
    brains = pc(UID=uid)
    if brains:
        return brains[0]


def make_brainGetter(context):
    """
    Return a function which looks up a UID
    and doesn't take a content argument;
    use this for multiple searches
    """
    pc = getToolByName(context, 'portal_catalog')._catalog

    def getbrain(uid):
        brains = pc(UID=uid)
        if brains:
            return brains[0]
    return getbrain


def make_pathByUIDGetter(context):
    """
    Return a function which looks up a UID
    and returns the path as stored in portal_catalog
    """
    pc = getToolByName(context, 'portal_catalog')._catalog
    indexes = pc.indexes

    def uid2path(uid):
        return indexes['path']._unindex[indexes['UID']._index[uid]]

    return uid2path


def make_timeformatter(context, **kwargs):
    util = getToolByName(context, 'translation_service')
    longFormat = kwargs.pop('longFormat', False)
    domain = kwargs.pop('domain', 'plonelocales')
    check_kwargs(kwargs)  # raises TypeError if necessary
    default_context = context
    func = util.ulocalized_time
    default_request = context.REQUEST

    def totime(time, longFormat=longFormat,
               context=None,
               domain='plonelocales',
               request=None):
        if not time:
            return None
        if context is None:
            context = default_context
        if request is None:
            request = default_request
        return func(time, longFormat,
                    context=context,
                    domain=domain,
                    request=request)

    return totime


def getPath(context, sep='/'):
    """
    Return the complete path from the Zope root.

    Without the sep option given, this is the same as the getPath method
    of Products.ZCatalog.CatalogPathAwareness.CatalogAware.
    """
    return sep.join(context.getPhysicalPath())


def parents(context):
    """
    Ersetzt den gleichnamigen Tomcom-Adapter
    """
    return context.REQUEST['PARENTS']


def parent_brains(context, parent=None, reverse=False):
    """
    Return the parents (as catalogued), as a list of brains

    parent -- the starting point (default: `context`)
    reverse -- return a reverted list? (default: False)
    """
    liz = []
    if not parent:
        if context.portal_type == 'Plone Site':
            return liz
        uid = context.UID()
        parent = getbrain(context, uid)

    while parent and parent.UID:
        liz.append(parent)
        parent = parent.getParent()
    if reverse:
        liz.reverse()
    return liz


def get_parent(context, **kwargs):
    """
    Return the nearest principia-folderish parent.

    Arguments:
      context -- the starting point

    Keyword-only:

      _as -- 'object' (default) or 'path', so far

      factory -- a function which takes an ancestor object
    """
    pop = kwargs.pop
    _as = pop('_as', None)
    func = pop('factory', None)
    if _as is not None:
        if func is not None:
            raise TypeError('Please specify *either* _as (%(_as)r) '
                            ' *or* factory (%(factory)r), '
                            'but not both!'
                            % locals())
        if _as == 'path':
            func = getPath
        elif _as == 'object':
            pass
        else:
            raise ValueError('Invalid _as spec %(_as)r'
                             % locals())
    check_kwargs(kwargs)  # raises TypeError if necessary
    for ancestor in context.REQUEST['PARENTS']:
        if getattr(ancestor, 'isPrincipiaFolderish', False):
            if func is None:
                return ancestor
            return func(ancestor)


def get_published_templateid(context):
    """
    Return the template id, as of the PUBLISHED variable, or None
    """
    request = context.REQUEST
    pu = request.get('PUBLISHED')
    if pu is None:
        return None
    try:
        return pu.__name__
    except AttributeError as e:
        logger = logging.getLogger('v.p.tools.context:get_published_templateid')
        logger.exception(e)
        return None



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
visaplan.plone.tools:setup._gs (GenericSetup usage)
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types

try:
    # Zope:
    from Products.CMFCore.utils import getToolByName

    # Plone:
    from plone.app.upgrade.utils import loadMigrationProfile
except ImportError:
    if __name__ != '__main__':  # for doctests
        raise


# visaplan:
from visaplan.tools.lands0 import make_default_prefixer

safe_context_id = make_default_prefixer('profile-',
                                        ['snapshot-'])


def extract_profile_name(kwargs):
    """
    Extract the profile argument from the given kwargs dict

    A little helper for the load_and_cook function:
    We require a profile argument for load_and_cook!

    We require a `profile` argument, and we'll use a `suffix` argument if
    given.

    First, a little test helper:
    >>> def epn(kw):
    ...     res = extract_profile_name(kw)
    ...     return res, kw
    >>> kw = {'profile': 'my.fancy.product:nondefault',
    ...       'tuples': [('cssregistry', 'portal_css')]}
    >>> epn(kw)                                # doctest: +NORMALIZE_WHITESPACE
    ('profile-my.fancy.product:nondefault',
     {'tuples': [('cssregistry', 'portal_css')]})

    Given a naked product name, a ':default' suffix is appended:
    >>> kw = {'profile': 'my.fancy.product'}
    >>> epn(kw)
    ('profile-my.fancy.product:default', {})
    >>> kw = {'profile': 'my.fancy.product', 'suffix': 'other'}
    >>> epn(kw)
    ('profile-my.fancy.product:other', {})

    An explicitly given suffix will override a suffix given in the profile
    argument:
    >>> kw = {'profile': 'my.fancy.product:default',
    ...       'suffix':  'another'}
    >>> epn(kw)
    ('profile-my.fancy.product:another', {})
    >>> kw = {'profile': 'snapshot-my.fancy.product:default',
    ...       'suffix':  'another'}
    >>> epn(kw)
    ('snapshot-my.fancy.product:another', {})

    """
    pop = kwargs.pop
    profile = pop('profile')
    profile = safe_context_id(profile)
    stem, sep, suff = profile.partition(':')
    suffix = pop('suffix', None)
    if suff:
        if suffix is not None:
            profile = stem + ':' + suffix
    else:
        if suffix is None:
            suffix = 'default'
        profile = stem + ':' + suffix
    return profile


def load_and_cook(context, logger, **kwargs):
    """
    Universal loader and cooker

    By default, we load and cook the Javascript and CSS resources (see the
    `tuples` option)
    of the given profile (profile, suffix options);

    but often we need to recompile (cook) those resources after processing
    foreign profile: This is what the `after_profiles` option is for.
    Here we currently expect complete specifications, including 'profile-' pre-
    and ':default' suffixes.
    """
    info = logger.info
    pop = kwargs.pop
    after_profiles = pop('after_profiles', None)
    if after_profiles is not None:
        if isinstance(after_profiles, six_string_types):
            after_profiles = after_profiles.split()
        for profile in after_profiles:
            info('Profile %(profile)r ...', locals())
            loadMigrationProfile(context, profile)

    tuples = pop('tuples', None)
    if tuples is None:
        tuples = [
            ('jsregistry',  'portal_javascripts'),
            ('cssregistry', 'portal_css'),
            ]

    profile = extract_profile_name(kwargs)
    setup = getToolByName(context, 'portal_setup')
    run_step = setup.runImportStepFromProfile
    for xmlname, toolname in tuples:
        info('Profile %(profile)r: %(xmlname)s[.xml]', locals())
        try:
            run_step(profile, xmlname)
        except Exception as e:
            logger.error('error running %(xmlname)r[.xml] from %(profile)s',
                         locals())
            logger.error('Exception: %(e)r', locals())
            logger.error('e.args: %s', (e.args,))
            raise

        tool = getToolByName(context, toolname)
        info('toolname %(toolname)r --> tool %(tool)r', locals())
        try:
            tool.cookResources()
        except AttributeError as e:
            logger.error('%(xmlname)s[.xml]: '
                         'The %(toolname)r tool lacks a cookResources method'
                         ' %(tool)r',
                         locals())
            raise
        else:
            info('%(xmlname)s resources cooked.', locals())


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: tw=79 cc=+1 sw=4 sts=4 et si
"""
visaplan.plone.tools: attools: Archetypes-related tools (_at4)
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types
from six import text_type as six_text_type

# 3rd party:
from bs4 import BeautifulSoup

# visaplan:
from visaplan.tools.coding import safe_decode
from visaplan.tools.html import from_plain_text

# Local imports:
from ._at3 import getter_name
from ._at5 import is_nonempty_string


def generate_all_texts(context, *fieldnames, **kwargs):
    """
    Generate all (existing) text fields as (fieldname, value) tuples

    Keyword options:

    - ignore_empty (default: True) --
        If a field exists but is empty (or contains a non-string value),
        try the next one;
        with ignore_empty=False,
        any value of the first existing field would be used.

    - decode (default: visaplan.plone.coding.safe_decode) --
        a function to decode a bytes string to unicode.
        Specify a falsy value to supporess decoding.

    If the field contains text/html (i.e., the content starts with
    '<'), it is converted to text/plain.
    """
    ignore_empty = kwargs.pop('ignore_empty', 1)
    if 'decode' not in kwargs:
        decode = safe_decode
    else:
        decode = kwargs.pop('decode')
        if not decode:
            decode = None

    if 'schema' in kwargs:
        schema = kwargs.pop('schema')
    elif fieldnames and not isinstance(fieldnames[0], six_string_types):
        fieldnames = list(fieldnames)
        schema = fieldnames.pop(0)
    else:
        schema = context.Schema()
    if schema is not None:
        raise ValueError('Sorry; using a dedicated schema here (%(schema)r)'
                         'is not yet supported'
                         % locals())

    if kwargs:
        raise TypeError('Unsupported keyword option(s): %s' % (
                        ', '.join(sorted(set(kwargs))),
                        ))
    if not fieldnames:
        fieldnames = [
                'title',
                'subject',
                'description',
                'text',
                ]
    for fieldname in fieldnames:
        try:
            val = getattr(context, fieldname)
        except AttributeError:
            gname = getter_name(fieldname)
            try:
                getter = getattr(context, gname)
            except AttributeError:
                continue
            else:
                val = getter()

        if isinstance(val, (list, tuple)):
            val = u' '.join([
                safe_decode(chunk)
                for chunk in val
                if is_nonempty_string(chunk)
                ])
        elif not isinstance(val, six_string_types):
            if ignore_empty:
                continue
            else:
                yield fieldname, ''
        if decode is not None and not isinstance(val, six_text_type):
            val = decode(val)
        if not val.startswith(u'<'):
            val = val.strip()
        if not val and ignore_empty:
            continue
        if val.startswith(u'<'):
            soup = BeautifulSoup(val, 'lxml')
            val = soup.text
        yield fieldname, val


def get_all_texts(context, *fieldnames, **kwargs):
    """
    Get all (existing) texts fields, joined to one unicode string
    """
    res = []
    for fieldname, value in generate_all_texts(context,
                                               *fieldnames, **kwargs):
        if value:
            res.append(value)
    return u' '.join(res)



# -*- coding: utf-8 -*-
"""
ZCML-Generierung für Plone-Ressourcen

Verwendung:
    from visaplan.plone.tools.zcmlgen import TemplateGenerator
    TemplateGenerator(__file__).write()

TO DO:
- Bislang haben wir spezialisierte Generatoren, die nur jeweils eine Sorte von
  Einträgen erzeugen.
  Wünschenswert wäre:
  - eine Generierung aller unterstützten Typen durch einen allgemeinen
    Generator
  - eine Gruppierung nach Typen
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types

# Standard library:
import sys
from difflib import Differ
from os import listdir
from os.path import abspath, isdir, isfile, join, normcase, sep, split

# Logging / Debugging:
from pdb import set_trace

try:
    # Zope:
    from Globals import DevelopmentMode
except ImportError:
    # Hotfix for Zope 4; how to properly replace this?
    if __name__ == '__main__':
        DevelopmentMode = True
    else:
        DevelopmentMode = False

# visaplan:
from visaplan.plone.base.exceptions import MissingFileError, OutdatedFileError

__all__ = [
        'BasicGenerator',
        'TemplateGenerator',
        'ResourceGenerator',
        'SubpackageGenerator',
        # Hilfsfunktionen:
        'make_plonename_factory',
        'changeable_path',
        ]


# -------------------------------------------- [ Daten ... [
_FRAME = """<configure
    xmlns="http://namespaces.zope.org/zope"
    xmlns:browser="http://namespaces.zope.org/browser">
    %s
</configure>
"""

RESOURCE = """
    <browser:resource
             name="%(filename)s"
             file="%(filename)s"
             />
"""

SUBPACKAGE = """
    <include package=".%(name)s" />
"""

PAGE = """
    <browser:page
             for="*"
             name="%(name)s"
             template="%(filename)s"
             permission="zope2.View"
             />
"""

_DENIED = frozenset(['__init__.py',
                     'configure.zcml',
                     '.svn'])
# -------------------------------------------- ] ... Daten ]


class BasicGenerator(object):
    """
    Abstrakte Basisklasse;
    fehlende Methoden z. B. in ResourceGenerator definiert
    """
    def __init__(self, fn, **kwargs):
        """
        Construct a BasicGenerator

        Arguments:

          fn -- the file name of the calling module (__file__)

        Options (keyword-only):

          skip -- names of entries to exclude from the generation.
                  A sequence; if a string, will be split.
        """
        dirname, filename = split(fn)
        self.base_path = abspath(dirname)
        self.blacklist = set(_DENIED)
        # die generierende Datei bleibt außen vor:
        self.blacklist.add(filename)
        skip_more = kwargs.pop('skip', None)
        if skip_more:
            if isinstance(skip_more, six_string_types):
                skip_more = skip_more.split()
            self.blacklist.update(skip_more)
        if kwargs:
            raise TypeError('Unknown arguments!')

        self.zcml_name = join(self.base_path, 'configure.zcml')

    def __repr__(self):
        return '<%s(%s/)>' % (self.__class__.__name__, self.base_path)

    def skip(self, fn, skipdirs=True):
        """
        Soll die aktuelle Datei übergangen werden?

        fn -- der Datei- (oder Verzeichnis-) -name
        skipdirs -- wenn True, werden Verzeichnisse generell ignoriert;
                    wenn False, werden Nicht-Verzeichnisse ignoriert
                    (SubpackageGenerator)
        """
        if fn in self.blacklist:
            return True
        elif fn.startswith('.'):
            return True
        elif fn.endswith('.pyc'):
            return True
        elif fn.endswith('.pyo'):
            return True
        elif fn.endswith('~'):
            return True
        elif fn.endswith('.orig'):
            return True
        elif isdir(join(self.base_path, fn)):
            return skipdirs
        else:
            return not skipdirs

    def generate_list(self):
        tmpls = []
        ENTRY_MASK = self._entry_mask()
        for dict_ in self.generate_dicts():
            tmpls.append(ENTRY_MASK % dict_)
        return tmpls

    def generate_text(self):
        return _FRAME % ''.join(self.generate_list())

    def write(self, devmode=None):
        """
        Schreibe eine aktuelle (generische) configure.zcml-Datei
        und wirf eine Exception, wenn sie hinterher nicht in der erwarteten Form existiert.

        Achtung - es findet ein relativ einfacher Textvergleich statt;
        sobald z. B. Anpassungen für einzelne Browserseiten nötig werden,
        ist diese Generierung nicht mehr anwendbar und muß deaktiviert werden!
        """
        txt = self.generate_text()
        fn = self.zcml_name
        found = isfile(fn)
        if found:
            with open(fn, 'ru') as fp:
                old = fp.read()
                if self.diff(old, txt):
                    return
        if devmode is None:
            devmode = DevelopmentMode
        if not changeable_path(fn, devmode):
            if found:
                raise OutdatedFileError(filename=fn)
            else:
                raise MissingFileError(filename=fn)
        with open(fn, 'w') as fp:
            fp.write(txt)

    def diff(self, old, new):
        """
        Gib (in Anlehnung an das Kommandozeilen-Tool) True zurück,
        wenn nichts geändert; ansonsten False.

        Ignoriere Leerzeilen.
        """
        def skipempty(s):
            return not s.rstrip()
        differ = Differ(linejunk=skipempty)
        oldl = [line.strip()
                for line in old.splitlines()
                ]
        newl = [line.strip()
                for line in new.splitlines()
                ]
        # a Generator object is "True", even if yielding an empty list ...
        difflines = list(differ.compare(oldl, newl))
        # Differ creates a human-readable difference; lines prefixed by '  '
        # are common to both
        if [difference
            for difference in difflines
            if not difference.startswith('  ')  # unchanged lines
            and difference not in ('+ ', '- ')  # empty lines added/removed
            ]:
            print('\n'.join(difflines))
            if sys.stdout.isatty(): set_trace()
            return False
        return True


class ResourceGenerator(BasicGenerator):
    """
    implementiert die fehlenden Methoden der BasicGenerator-Klasse
    für statische Ressourcen
    """

    def generate_dicts(self):
        for fn in sorted(listdir(self.base_path)):
            if self.skip(fn):
                continue
            dic = {'filename': fn,
                   'name': fn,
                   }
            yield dic

    def _entry_mask(self):
        return RESOURCE


class SubpackageGenerator(BasicGenerator):
    """
    implementiert die fehlenden Methoden der BasicGenerator-Klasse
    für Unterpakete
    """

    def generate_dicts(self):
        for fn in sorted(listdir(self.base_path)):
            if self.skip(fn, skipdirs=False):
                continue
            fullname = join(self.base_path, fn)
            considered_package = False
            # when we find one, we expect the other to be present as well!
            # If either is missing, there will be (most likely) some error,
            # but this is not to be solved *here*.
            for pfn in ('__init__.py',
                        'configure.zcml',
                        ):
                if isfile(join(fullname, pfn)):
                    considered_package = True
                    break
            if considered_package:
                yield {'name': fn,
                       }

    def _entry_mask(self):
        return SUBPACKAGE


def make_plonename_factory(ext):
    """
    Erzeugt eine Funktion, die für einen gegebenen Dateinamen
    den aufrufbaren Namen im Plone-Kontext zurückgibt.

    >>> f = make_plonename_factory('.pt')
    >>> f.__name__
    'make_plonename.pt'
    >>> f('test.py')
    >>> f('test.pt')
    'test'
    """
    if not ext.startswith('.'):
        raise ValueError('.ext erwartet (%(ext)r erhalten)'
                         % locals())
    lext = len(ext)

    def make_plonename(fn):
        if fn.endswith(ext):
            return fn[:-lext] or None

    make_plonename.__name__ += ext
    return make_plonename


class TemplateGenerator(BasicGenerator):
    """
    Generator für Templates
    """

    def __init__(self, fn, aliases=(), aliasdict_func=None,
                 include_extensions=('.xml.pt', '.pt',
                                     ), **kwargs):
        """
        aliases-Funktionalität derzeit noch nicht in Verwendung
        """
        BasicGenerator.__init__(self, fn, **kwargs)

        if aliases:
            if isinstance(aliases, set):
                aliases = sorted(aliases)
            self.aliases = aliases
        else:
            self.aliases = None
        self.aliasdict_func = aliasdict_func
        self.whitelist_functions = [make_plonename_factory(ext)
                                    for ext in include_extensions
                                    ]

    def _entry_mask(self):
        return PAGE

    def generate_dicts(self):
        whitelist_functions = self.whitelist_functions
        for fn in sorted(listdir(self.base_path)):
            if self.skip(fn):
                continue
            pn = None  # Plone-Name
            for f in whitelist_functions:
                pn = f(fn)
                if pn is not None:
                    break
            if pn is None:
                print('SKIPPING: %s' % join(self.base_path, fn))
                continue
            yield {'filename': fn,
                   'name': pn,
                   }

    def generate_list(self):
        tmpls = BasicGenerator.generate_list(self)
        if self.aliases:
            ENTRY_MASK = self._entry_mask()
            first_nondict = True
            func = None
            for alias in aliases:
                if not isinstance(alias, dict):
                    if first_nondict:
                        assert self.aliasdict_func is not None
                        assert callable(self.aliasdict_func)
                        func = self.aliasdict_func
                        first_nondict = False
                    alias = func(alias)
                tmpls.append(ENTRY_MASK % alias)
        return tmpls


def changeable_path(fullpath, devmode=None):
    """
    Ist der übergebene Pfad einer, in dem wir eine Datei zu ändern versuchen?

    Im empfohlenen Plone-Setup hat der laufende Zope-Prozeß keine Schreibrechte
    in den Programmquellen.  Zur Entwicklung ist es dennoch praktisch,
    bestimmte Dateien automatisch generieren zu können, z. B.
    Versionsinformationen oder einfache configure.zcml-Dateien.

    Die vorgeschlagene Logik ist wie folgt:

    1. Prüfen, ob die Zieldatei schon den erwarteten Wert enthält
       (nicht Bestandteil dieser Funktion).
       Wenn ja, dann sind weitere Aktionen unnötig.
    2. Prüfen, ob die fragliche Datei nach ihrem Namen und ggf. dem übergebenen
       globalen Entwicklungsmodus änderbar ist
       (diese Funktion).
       Wenn nein: eine Ausnahme werfen, um den Start der Instanz zu verhindern,
       damit dieses Faktum nicht unbemerkt bleibt!
    3. Die Änderung versuchen.
       Dabei kann natürlich dennoch ein OSError auftreten; das wäre dann evtl.
       ein Anlaß, die Vorab-Prüfung zu verfeinern.

    Dateien mit den Namen 'version.txt' oder 'configure.zcml' dürfen in
    Entwicklungspaketen geändert werden:
    >>> changeable_path('.../src/visaplan.cool.tool/src/.../version.txt')
    True
    >>> changeable_path('.../src/visaplan.cool.tool/src/.../configure.zcml')
    True

    Andere Dateinamen werden zurückgewiesen:
    >>> changeable_path('.../src/visaplan.cool.tool/src/.../otherfile.txt')
    False

    In Eggs wird generell nichts geändert:
    >>> changeable_path('.../eggs/visaplan.cool.tool-1.0-py2.7.egg/src/.../configure.zcml')
    False
    """
    splitpath = normcase(abspath(fullpath)).split(sep)
    filename = splitpath.pop()
    # TODO: normcase for whitelist
    if filename not in ('version.txt',
                        'configure.zcml',
                        ):
        return False
    for seg in splitpath:
        if seg.endswith('.egg'):
            return False
    if not set(['src', 'Products']).intersection(splitpath):
        return False
    if devmode is None:
        devmode = DevelopmentMode
    return devmode or False


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
There is currently absolutely no point in "activation" of this package as an add-on;
since we did so before, we try to support clean deactivation.

So, please DON'T activate!
We try our best not to trigger you to do so.
"""
# Python compatibility:
from __future__ import absolute_import

__author__ = """Tobias Herp <tobias.herp@visaplan.com>"""
__docformat__ = 'plaintext'

# Zope:
from zope.interface import implementer

# Plone:
from Products.CMFPlone.interfaces import INonInstallable

__all__ = [
    'HiddenProfiles',
    ]

# ------------------------------------------------------ [ Daten ... [
PROJECTNAME = 'visaplan.plone.tools'
PROFILE_ID = PROJECTNAME + ':default'
LOGGER_LABEL = PROJECTNAME + ': setuphandlers'
# ------------------------------------------------------ ] ... Daten ]


@implementer(INonInstallable)
class HiddenProfiles(object):

    def getNonInstallableProfiles(self):
        """Hide uninstall profile from site-creation and quickinstaller."""
        return [
            PROJECTNAME + ':uninstall',
        ]


@implementer(INonInstallable)
class HiddenProducts(object):

    def getNonInstallableProducts(self):
        return [PROJECTNAME]


def post_install(context):
    """Post install script"""
    # Do something at the end of the installation of this package.


def uninstall(context):
    """Uninstall script"""
    # Do something at the end of the uninstallation of this package.




# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# Python compatibility:
from __future__ import absolute_import, print_function

from six import iteritems as six_iteritems
from six import text_type as six_text_type

__all__ = [
    'split_group_id',
    'build_groups_set',
    ],

try:
    # visaplan:
    from visaplan.tools.coding import safe_decode
except ImportError:
    if __name__ != '__main__':
        raise
    def safe_decode(s):
        if isinstance(s, six_text_type):
            return s
        return s.decode('utf-8')

# ------------------------------------------------------- [ data ... [
# (from visaplan.plone.groups.unitraccgroups.utils)
# Die Reihenfolge ist wichtig!
STRUCTURE_GROUP_SUFFIXES = ['Reader', 'Author']

try:
    # visaplan:
    from visaplan.tools.classes import connected_dicts
except ImportError:
    if __name__ == '__main__':
        print('Dictionaries are not connected')
    else:
        raise
    _STRUCTURE_SUFFIX2ROLE = dict({
        'Author': 'Editor',
        'Reader': 'Reader',
    })
    _STRUCTURE_ROLE2SUFFIX = {}
    for key, val in _STRUCTURE_SUFFIX2ROLE.items():
        _STRUCTURE_ROLE2SUFFIX[val] = key
else:
    _STRUCTURE_SUFFIX2ROLE, _STRUCTURE_ROLE2SUFFIX = connected_dicts(**{
    'Author': 'Editor',
    'Reader': 'Reader',
    })
SUFFIX2ROLE = dict(_STRUCTURE_SUFFIX2ROLE)
SUFFIX2ROLE.update({
    'learner': 'Reader',
    # Die Alumni-Gruppe vermittelt keine Rolle
    # direkt auf dem betroffenen Objekt:
    'alumni': None,
    })
# -------------------------------- [ Gruppen für Kurse ... [
# Achtung, Kleinschreibung:
LEARNER_SUFFIX = intern('learner')
ALUMNI_SUFFIX = intern('alumni')
COURSE_GROUP_SUFFIXES = [LEARNER_SUFFIX, ALUMNI_SUFFIX]
# -------------------------------- ] ... Gruppen für Kurse ]
ALL_GROUP_SUFFIXES = STRUCTURE_GROUP_SUFFIXES + COURSE_GROUP_SUFFIXES
UID_CHARS = frozenset('0123456789abcdef')
SIMPLE_GROUP_INFO = {
        'uid': None,
        # das (z B. für 'Author') von der Rolle verschiedene *Suffix*:
        'role': None,
        }
RESOLVED_GROUP_INFO = {
        'uid': None,
        'suffix': None,
        'role': None,  # hier die real zuzuordnende Rolle
        }
PRETTY_MASK = {}
for role in ALL_GROUP_SUFFIXES:
    PRETTY_MASK[role] = u'%s group "{group}"' % role
if 0 and 'nur fuer den Parser':
    _('Author group "{group}"')
    _('Reader group "{group}"')
    _('learner group "{group}"')
    _('alumni group "{group}"')
# ------------------------------------------------------- ] ... data ]


def split_group_id(gid, resolve_role=False):
    """
    NOTE: This function is currently still quite specific;
          we might want to make it more general one day,
          e.g. using some more arguments.

    Splitte die übergebene Gruppen-ID auf und gib ein Dictionary zurück,
    das über eine etwaige "Rollenkomponente" informiert und dann auch
    die uid des zugehörigen Objekts enthält.

    Achtung:
        Der Schlüssel 'role' enthält bisher nicht wirklich den Namen der
        Rolle, sondern das Suffix, das von diesem in wichtigen Fällen abweicht!
        Um dies zu beheben, resolve_role=True übergeben;
        dies schreibt nach 'role' die *real zu verwendende* Rolle,
        und das Suffix steht im zusätzlichen Schlüssel 'suffix'.

        In einer späteren Version wird der Vorgabewert von resolve_role auf
        True geändert, und noch etwas später die Unterstützung für False
        entfernt.

    Zunächst die Tests für resolve_role=False:

    Die Schlüssel 'uid' und 'role' haben String-Werte,
    sofern der Mittelteil wie eine korrekte UID aussieht und das Suffix
    eine der hierfür bekannten Rollen ist;
    ansonsten sind sie None.

    >>> split_group_id('group_f6350ab731c3601e925eac482206bda5_Author')
    {'role': 'Author', 'uid': 'f6350ab731c3601e925eac482206bda5'}

    Bei "normalen" Gruppen hat die scheinbare (!) UID keine spezielle
    Bedeutung und wird daher nicht als solche zurückgegeben:

    >>> split_group_id('group_0123456789abcdef0123456789abcdef')
    {'role': None, 'uid': None}

    Das Suffix 'learner' repräsentiert keine echte Rolle (es wird stattdessen
    'Reader' vermittelt); es ist aber ein bekanntes Suffix
    und wird (mit resolve_role=False) als 'role' zurückgegeben
    (mit resolve_role=True als 'suffix'; siehe unten):

    >>> split_group_id('group_f6350ab731c3601e925eac482206bda5_learner')
    {'role': 'learner', 'uid': 'f6350ab731c3601e925eac482206bda5'}

    Wenn Unicode übergeben wird, kommt Unicode zurück:

    >>> split_group_id(u'group_f6350ab731c3601e925eac482206bda5_Author')
    {'role': u'Author', 'uid': u'f6350ab731c3601e925eac482206bda5'}

    Es wird genau auf Plausibilität geprüft und ggf. ein Dict. mit
    None-Werten zurückgegeben.

    Falsche Zeichen in der UID:
    >>> split_group_id('group_FOOBARBAZ1c3601e925eac482206bda5_Author')
    {'role': None, 'uid': None}

    Unbekanntes oder leeres Suffix:
    >>> split_group_id('group_f6350ab731c3601e925eac482206bda5_')
    {'role': None, 'uid': None}
    >>> split_group_id('group_f6350ab731c3601e925eac482206bda5_Foo')
    {'role': None, 'uid': None}

    Falsche Länge der UID:
    >>> split_group_id('group_f6350ab731c3601c482206bda5_Author')
    {'role': None, 'uid': None}

    Falsches Präfix:
    >>> split_group_id('gröup_f6350ab731c3601e925eac482206bda5_learner')
    {'role': None, 'uid': None}
    >>> split_group_id(u'gröup_f6350ab731c3601e925eac482206bda5_learner')
    {'role': None, 'uid': None}

    Nun ergänzend noch Tests für resolve_role=True:

    >>> def sgi(gid):
    ...     return sorted(split_group_id(gid, resolve_role=True).items())
    >>> sgi('group_0123456789abcdef0123456789abcdef')
    [('role', None), ('suffix', None), ('uid', None)]

    >>> sgi('group_f6350ab731c3601e925eac482206bda5_Reader')
    [('role', 'Reader'), ('suffix', 'Reader'), ('uid', 'f6350ab731c3601e925eac482206bda5')]
    >>> sgi('group_f6350ab731c3601e925eac482206bda5_Author')
    [('role', 'Editor'), ('suffix', 'Author'), ('uid', 'f6350ab731c3601e925eac482206bda5')]
    >>> sgi('group_f6350ab731c3601e925eac482206bda5_learner')
    [('role', 'Reader'), ('suffix', 'learner'), ('uid', 'f6350ab731c3601e925eac482206bda5')]

    Die Alumni-Gruppe vermittelt keine Rolle direkt auf dem betroffenen Objekt;
    daher ist die Rolle hier None:
    >>> sgi('group_f6350ab731c3601e925eac482206bda5_alumni')
    [('role', None), ('suffix', 'alumni'), ('uid', 'f6350ab731c3601e925eac482206bda5')]

    """
    liz = gid.split('_', 2)
    if resolve_role:
        res = dict(RESOLVED_GROUP_INFO)
    else:
        res = dict(SIMPLE_GROUP_INFO)

    if not liz[2:]:
        return res
    elif liz[0] != 'group':
        return res
    elif liz[2] not in ALL_GROUP_SUFFIXES:
        return res
    uid = liz[1]
    if not uid:
        return res
    elif set(uid).difference(UID_CHARS):
        return res
    elif len(uid) != 32:
        return res
    res['uid'] = uid
    if resolve_role:
        res['suffix'] = suffix = liz[2]
        res['role'] = SUFFIX2ROLE[suffix]
    else:
        res['role'] = liz[2]
    return res


def build_groups_set(dic, userid):
    """
    Hilfsfunktion für is_member_of_factory

    >>> dic = {'group_a': ['group_b', 'group_c'],
    ...        'group_b': ['user_a', 'user_b'],
    ...        'group_c': ['user_c'],
    ...        }
    >>> groups = build_groups_set(dic, 'user_a')
    >>> sorted(groups)
    ['group_a', 'group_b', 'user_a']
    """
    groups = set([userid])
    _traverse_dict(dic, groups)
    return groups


def _traverse_dict(dic, groups):
    """
    Hilfsfunktion für build_groups_set:
    realisiert die Rekursion.
    Es wird abgebrochen, wenn es keine Änderungen mehr gab.

    >>> dic = {'group_a': ['group_b', 'group_c'],
    ...        'group_b': ['user_a', 'user_b'],
    ...        'group_c': ['user_c'],
    ...        }

    <groups> enthält zunächst die <uid> (oder gid) selbst:
    >>> groups = set(['user_a'])

    Der Rückgabewert wird üblicherweise nicht verwendet;
    er gibt die Anzahl der Iterationen an:
    >>> _traverse_dict(dic, groups)
    2
    >>> sorted(groups)
    ['group_a', 'group_b', 'user_a']
    """
    iterations = 1
    while True:
        finished = True
        for gid, members in six_iteritems(dic):
            if gid in groups:
                continue
            if groups.intersection(members):
                groups.add(gid)
                finished = False
        if finished:
            break
        iterations += 1
    return iterations


if __name__ == '__main__':
    # Standard library:
    from doctest import testmod
    testmod()



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
visaplan.plone.tools.env: Information about the Zope environment
"""

# Python compatibility:
from __future__ import absolute_import

# Standard library:
from os import environ
from os.path import basename, curdir, normpath, sep

__all__ = [
    'worker_name',    # the name of the worker, e.g. 'instance' or 'client1'
    'var_directory',  # the 'var' directory path of the instance
    'lockfiles_directory',  # LOCKFILES_DIR, or {var}/lock
    ]


def worker_name(_env=None):
    """
    Return the name of the worker, e.g. 'client1'

    For long-running processes, e.g. upgrade steps, we might want to know the
    worker process which executes them; that way we are able to exclude that
    worker temporarily from the load balancing.

    We accept the `_env` option for testability,
    and for calls by the var_directory function (below).

    >>> e = dict(INSTANCE_HOME='/opt/zope/parts/instance')
    >>> worker_name(e)
    'instance'
    """
    if _env is None:
        _env = environ
    home = _env.get('INSTANCE_HOME')
    if home:
        worker = basename(home)
        return worker or None


def var_directory(_env=None, strict=0):
    """
    Return the (full) path to the "var" directory of the instance.

    Well; as we see it.  There might be a more reliable method to get this
    in a present or future Plone release; we might integrate it according to
    knowledge and (our) need.

    We accept the `_env` option for testability only.

    The CLIENT_HOME is always available, right?

    >>> e = dict(CLIENT_HOME='/opt/zope/instances/myplone/var/instance',
    ...        INSTANCE_HOME='/oh/whatever/parts/instance')
    >>> var_directory(e)
    '/opt/zope/instances/myplone/var'

    You might have configured a VAR_ROOT variable in your buildout,
    which will take precedence:
    >>> e.update({'VAR_ROOT': '/some/var/root/elsewhere'})
    >>> var_directory(e)
    '/some/var/root/elsewhere'

    What if the environment doesn't contain anything useful?
    >>> e=dict(junk='only')

    By default, we currently tolerate this and return None:
    >>> var_directory(e) is None
    True

    But if you prefer, you can be strict as well:
    >>> var_directory(e, strict=1)
    Traceback (most recent call last):
      ...
    KeyError

    Some corner case: The CLIENT_HOME contains the name of the instance binary
    only:
    >>> e = dict(CLIENT_HOME='instance',
    ...        INSTANCE_HOME='/oh/whatever/parts/instance')
    >>> var_directory(e) is curdir
    True
    """
    if _env is None:
        _env = environ
    val = _env.get('VAR_ROOT')
    if val:
        return val
    wn = worker_name(_env)
    val = _env.get('CLIENT_HOME')
    if val:
        liz = list(normpath(val).split(sep))
        if wn is not None:
            try:
                idx = liz.index(wn)
            except ValueError:
                pass
            else:
                del liz[idx:]
        if liz:
            return sep.join(liz)
        else:
            return curdir
    if strict:
        raise KeyError


def lockfiles_directory(_env=None, strict=0):
    """
    Return the (full) path to the directory to contain the lock files.

    This is aimed to contain lock files as created by zc.lockfile.LockFile,
    e.g. using our ConvenientLock wrapper class in the .lock module.

    We take into account, by precendence:

    1. A LOCKFILES_DIR environment variable, if present
    2. var_directory() + '/lock'

    We accept the `_env` option for testability only.

    >>> e = dict(CLIENT_HOME='/opt/zope/instances/myplone/var/instance',
    ...        INSTANCE_HOME='/oh/whatever/parts/instance')
    >>> lockfiles_directory(e)
    '/opt/zope/instances/myplone/var/lock'

    You might have configured a VAR_ROOT variable in your buildout,
    which will take precedence:
    >>> e.update({'VAR_ROOT': '/some/var/root/elsewhere'})
    >>> lockfiles_directory(e)
    '/some/var/root/elsewhere/lock'

    A LOCKFILES_DIR value, of course, will always win:
    >>> e.update({'LOCKFILES_DIR': '/i/want/cows'})
    >>> lockfiles_directory(e)
    '/i/want/cows'

    What if the environment doesn't contain anything useful?
    >>> e=dict(junk='only')

    By default, we currently tolerate this and return None:
    >>> lockfiles_directory(e) is None
    True

    But if you prefer, you can be strict as well:
    >>> lockfiles_directory(e, strict=1)
    Traceback (most recent call last):
      ...
    KeyError

    What if we don't have a LOCKFILES_DIR variable, and
    the CLIENT_HOME contains the name of the instance binary only?
    >>> e = dict(CLIENT_HOME='instance',
    ...        INSTANCE_HOME='/oh/whatever/parts/instance')
    >>> lockfiles_directory(e).startswith(curdir)
    True
    >>> lockfiles_directory(e)
    './lock'
    """
    if _env is None:
        _env = environ
    val = _env.get('LOCKFILES_DIR')
    if val:
        return val
    val = var_directory(_env, strict)
    if val:
        return sep.join([normpath(val), 'lock'])
    elif strict:
        raise KeyError


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -------------------------------- [ groupinfo_factory ... [
# Python compatibility:
from __future__ import print_function


def groupinfo_factory(context, pretty=0, forlist=0, searchtext=0):
    """
    Factory-Funktion; als Ersatz für get_group_info_by_id, insbesondere für
    wiederholte Aufrufe.  Indirektionen etc. werden einmalig beim Erstellen der
    Funktion aufgelöst und anschließend in einer Closure vorgehalten.

    pretty -- Gruppennamen auflösen, wenn Suffix an ID und Titel
    forlist -- Minimale Rückgabe (nur ID und Titel),
               aber mit pretty kombinierbar
    """
    pg = getToolByName(context, 'portal_groups')
    get_group = pg.getGroupById
    acl = getToolByName(context, 'acl_users')
    translate = make_translator(context)
    GROUPS = acl.source_groups._groups

    def minimal_group_info(group_id):
        """
        Nur ID und Titel
        """
        group = get_group(group_id)

        dict_ = {'id': group_id}
        try:
            thegroup = GROUPS[group_id]
            dict_['group_title'] = safe_decode(thegroup['title'])
            return dict_
        except KeyError:
            return {}

    def basic_group_info(group_id):
        """
        Gib ein Dict. zurück;
        - immer vorhandene Schlüssel:
          id, group_title, group_description, group_manager, group_desktop
        - nur bei automatisch erzeugten Gruppen: role, role_translation, brain

        Argumente:
        group_id -- ein String, normalerweise mit 'group_' beginnend
        """

        group = get_group(group_id)

        dict_ = {'id': group_id}
        try:
            thegroup = GROUPS[group_id]
            dict_['group_title'] = safe_decode(thegroup['title'])
        except KeyError:
            return {}
        dict_['group_description'] = safe_decode(thegroup['description'])
        dict_['group_manager'] = thegroup.get('group_manager')
        dict_['group_desktop'] = thegroup.get('group_desktop')

        dic = split_group_id(group_id)
        if dic['role'] is not None:
            dict_.update(dic)

        dict_['role_translation'] = translate(dic['role'])
        dict_['brain'] = getbrain(context, dic['uid'])
        return dict_

    def pretty_group_info(group_id):
        """
        Ruft basic_group_info auf und fügt einen Schlüssel 'pretty_title'
        hinzu, der den Gruppentitel ohne das Rollensuffix enthält.
        """
        dic = basic_group_info(group_id)
        try:
            if 'role' not in dic:
                dic['pretty_title'] = translate(dic['group_title'])
                return dic
            liz = dic['group_title'].split()
            if liz and liz[-1] == dic['role']:
                stem = u' '.join(liz[:-1])
                mask = PRETTY_MASK[dic['role']]
                dic['pretty_title'] = translate(mask).format(group=stem)
            else:
                dic['pretty_title'] = translate(dic['group_title'])
        except KeyError as e:
            # evtl. keine Gruppe! (Aufruf durch get_mapping_group)
            print(e)
            pass
        return dic

    def minimal2_group_info(group_id):
        """
        Ruft minimal_group_info auf und modifiziert ggf. den Schlüssel
        'group_title' (entsprechend dem von pretty_group_info zurückgegebenen
        Schlüssel 'pretty_title')
        """
        dic = minimal_group_info(group_id)
        if not dic:     # Gruppe nicht (mehr) gefunden;
            return dic  # {} zurückgeben
        dic2 = split_group_id(group_id)
        if dic2['role'] is not None:
            dic.update(dic2)
        if 'role' not in dic:
            dic['group_title'] = translate(dic['group_title'])
            return dic
        liz = dic['group_title'].split()
        if liz and liz[-1] == six_text_type(dic['role']):
            stem = u' '.join(liz[:-1])
            mask = PRETTY_MASK[dic['role']]
            dic['group_title'] = translate(mask).format(group=stem)
        else:
            dic['group_title'] = translate(dic['group_title'])
        return dic

    def make_searchstring(group_info):
        """
        Arbeite direkt auf einem group_info-Dict;
        Gib kein Dict zurück, sondern einen String für Suchzwecke
        """
        try:
            group_id = group_info['id']
        except KeyError:
            raise
        try:
            res = [safe_decode(group_info['title']),
                   safe_decode(group_id)]
        except KeyError:
            print((list(group_info.items())))
            raise
            return u''
        dic2 = split_group_id(group_id)
        prettify = True
        for val in dic2.values():
            if val is None:
                prettify = False
                break  # es sind immer alle None, oder alle nicht None
            else:
                res.append(safe_decode(val))
        if prettify:
            pretty = pretty_group_title(group_id, res[0], translate)
            if pretty is not None:
                res.append(safe_decode(pretty))
        descr = group_info['description']
        if descr:
            res.append(safe_decode(descr))
        return u' '.join(res)

    if searchtext:
        return make_searchstring
    if forlist:
        if pretty:
            return minimal2_group_info
        else:
            return minimal_group_info
    if pretty:
        return pretty_group_info
    else:
        return basic_group_info
# -------------------------------- ] ... groupinfo_factory ]



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
dxtools - Dexterity tools

Dexterity should be installed to use this module;
otherwise, some functions might fail
"""

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Local imports:
from ._dx4 import get_first_text_as_html
from ._dx5 import generate_all_texts, get_all_texts



# Python compatibility:
from __future__ import absolute_import

import six

is_encoded = isinstance(s, str)  # Python 2: str
is_unicode = isinstance(s, six.text_type)  # Python 2: unicode
is_anystring = isinstance(s, six.string_types)  # Python 2: basestring
is_bytes = isinstance(s, bytes)



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Object tools for (_get_object).make_object_getter()
and (_make_folder).make_subfolder_creator()

All tools in this module are called in the same style.
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types
from six.moves import map

# Standard library:
from collections import defaultdict

# Zope:
from Products.CMFCore.utils import getToolByName

# Local imports:
from visaplan.plone.tools._have import HAS_SUBPORTALS, HAS_VPSEARCH
from visaplan.plone.tools.setup._args import (
    _extract_move_args,
    extract_layout_switch,
    extract_menu_switch,
    )
from visaplan.plone.tools.setup._exc import \
    AlreadyTranslated  # from LinguaPlone, or a dummy
from visaplan.plone.tools.setup._exc import \
    CantAddTranslationReference  # ... enhanced information

# Logging / Debugging:
from pdb import set_trace
from visaplan.tools.debug import pp

__all__ = [
        'handle_title',
        'handle_language',
        'handle_layout',
        'handle_menu',
        # minilogging system:
        'make_miniloggers',  # used here
        'make_notes_logger', # used in calling functions
        # operating on connected objects:
        ]

if HAS_SUBPORTALS:
    __all__.extend([
        'might_set_subportal',
        'handle_subportal',
        ])

if HAS_VPSEARCH:
    __all__.extend([
        'handle_united_search',
        ])
    # visaplan:
    from visaplan.plone.search.utils import (
        expand_localsearch,
        normalize_localsearch,
        storable_localsearch,
        )

# ----------------------------------------- [ minilogging system ... [
def make_notes_logger(logger, append_to=None):
    func = defaultdict(lambda: logger.error)
    func.update({
        'INFO': logger.info,
        'DEBUG': logger.debug,
        'ERROR': logger.error,
        })

    def logtup(tup):
        key, txt = tup
        func[key](txt)
        if append_to:
            append_to.append(tup)

    return logtup


# allows to write full information to a list
def make_miniloggers(notes, keys=None):
    """
    Make a few functions!
    """
    if not keys:
        keys = 'DEBUG INFO ERROR'.split()
    def makelogfunc(key):
        def func(txt, mapping=None):
            if mapping is not None:
                if (isinstance(mapping, dict)
                    and 'o' in mapping
                    and '%(o)r' not in txt
                    ):
                    txt = '(%(o)r) '+txt
                txt %= mapping
            notes.append((key, txt))
        func.__name__ = key.lower()
        return func

    return tuple(map(makelogfunc, keys))
# ----------------------------------------- ] ... minilogging system ]


def handle_title(o, kwdict, created, do_pop=True):
    """
    title -- a given (chosen) title
    default_title -- a calculated title (computed from an id)
    set_title -- set it?
    """
    changed = False
    notes = []
    _DBG, _NFO, _ERR = make_miniloggers(notes)

    pop = (kwdict.pop if do_pop
           else kwdict.get)

    title_given =   'title' in kwdict
    title =         pop('title', None)
    default_title = pop('default_title', None)
    if title:
        title = title.strip()
    if not title:
        title = default_title
    set_title =     pop('set_title', None)
    if created:
        found_title = 0
    else:
        found_title = o.Title()

    if set_title is None:
        if created:
            set_title = bool(title or default_title)
        else:
            set_title = not found_title and bool(title or default_title)

    if title:
        if title != found_title:
            if set_title and title:
                o.setTitle(title)
                _NFO('%(o)r: old title %(found_title)r --> new title %(title)r',
                      locals())
                changed = True
            elif title:
                _ERR('%(o)r: old title %(found_title)r mismatches %(title)r',
                     locals())
            elif set_title >= 2:
                if title is None:
                    _DBG('%(o)r: to remove the old title %(found_title)r, provide the empty string',
                         locals())
                else:
                    _NFO('%(o)r: resetting old title %(found_title)r to the empty string',
                         locals())
                    o.setTitle('')
            elif set_title:
                _DBG('%(o)r: to remove the old title %(found_title)r, '
                     "specify title='' and set_title=2", locals())
        else:
            _NFO('%(o)r: checked title (%(found_title)r)', locals())
    elif set_title >= 2 and title_given:
        _NFO('%(o)r: resetting old title %(found_title)r to the empty string',
             locals())
        o.setTitle('')
    elif set_title:
        _DBG('%(o)r: no title given', locals())
        if found_title:
            _DBG('%(o)r: to remove the old title %(found_title)r, '
                 "specify title='' and set_title=2", locals())
    return changed, notes


def handle_language(o, kwdict, created, do_pop=True):  # ----- [ h.l. ... [
    """
    Handle the language and canonical keyword arguments for the given object
    `o`.

    o -- the object
    kwdict -- the kwargs dictionary of the calling function (will be consumed)
    created -- a boolean to tell whether the given object has just been
               created, affecting a few defaults

    Return a 2-Tuple (changed, notes);
    `notes` is in turn a list of ('INFO', `text`) tuples.
    """
    changed = False
    notes = []
    _DBG, _NFO, _ERR = make_miniloggers(notes)
    pt = getattr(o, 'portal_type', None)
    if pt == 'Plone Site':
        _NFO("%(o)r: Won't handle_language for site root!" % locals())
        return changed, notes
    elif 0:
        pp(o, kwdict, pt)
        set_trace()

    pop = (kwdict.pop if do_pop
           else kwdict.get)

    language_given =  'language' in kwdict
    language =        pop('language', None)
    set_language =    pop('set_language', None)
    if language_given and not created:
        found_language = o.Language()
    else:
        found_language = 0

    canonical_given = 'canonical' in kwdict
    canonical =       pop('canonical', None)
    set_canonical =   pop('set_canonical', None)
    if canonical_given and not created:
        found_canonical = o.getCanonical()
    else:
        found_canonical = 0

    if language_given:
        if set_language is None:
            _DBG('set_language?')
            if created:
                set_language = bool(language)
                _DBG('%(set_language)r (freshly created'
                     ' with language=%(language)r', locals())
            elif language != found_language:
                set_language = True
                _DBG('%(set_language)r (old and new values different;'
                     '%(found_language)r --> %(language)r)', locals())
            else:
                set_language = False
                _DBG('%(set_language)r (found no reason;'
                     '%(found_language)r --> %(language)r,'
                     '%(found_canonical)r --> %(canonical)r)', locals())

        if set_language is None:
            set_language = language != found_language

    if canonical:
        if created:
            if not language:
                _ERR('canonical given %(canonical)r but no language %(language)r'
                     ' for a freshly created object!', locals())
                set_canonical = False
        else:
            if not language_given:
                _DBG('canonical given %(canonical)r but no language; '
                     'using %(found_language)r ...', locals())
                language = found_language

    # ----------------------- [ canonical, with language implied ... [
    if canonical_given:
        if set_canonical is None:
            _DBG('set_canonical?')
            if not canonical and not found_canonical:
                set_canonical = False
                _DBG('%(set_canonical)r (old and new values falsy;'
                     '%(found_canonical)r, %(canonical)r)', locals())
            elif canonical != found_canonical:
                _DBG('%(set_canonical)r (different values;'
                     '%(found_canonical)r --> %(canonical)r)', locals())
                can_path = '(unknown)'
                found_path = '(unknown)'
                try:
                    tup = canonical.getPhysicalPath()
                    can_path = '/'.join(tup)
                    tup = found_canonical.getPhysicalPath()
                    found_path = '/'.join(tup)
                    set_canonical = found_path != can_path
                except Exception as e:
                    _ERR('Error! (%(e)r)', locals())
                    raise
                finally:
                    _NFO('\n<new canonical>.getPath = %(can_path)r,'
                         '\n<old canonical>.getPath = %(found_path)r', locals())
                if not set_canonical:
                    _DBG('"different" objects but equal paths!')
                    set_trace()
            elif (found_canonical
                    and language_given
                    and language != found_language):
                set_canonical = True
                _DBG('%(set_canonical)r (language change;'
                     '%(found_language)r --> %(language)r, '
                     '%(found_canonical)r --> %(canonical)r)', locals())
            else:
                set_canonical = False
                _DBG('%(set_canonical)r (found no reason;'
                     '%(found_language)r --> %(language)r (%(set_language)r),'
                     '%(found_canonical)r --> %(canonical)r)', locals())

        if set_canonical and not language:
            _ERR('canonical given (%(canonical)r) but no language!'
                 ' (%(language)r)', locals())
            set_canonical = False

        if set_canonical:
            if found_canonical and not created:
                _NFO('Resetting language (was %(found_language)r, '
                     'canonical was %(found_canonical)r)', locals())
                o.setLanguage('')
                changed = True
            if language:
                _NFO('Setting language to %(language)r', locals())
                o.setLanguage(language)
                changed = True
            if canonical and canonical != o:
                _NFO('Setting canonical to %(canonical)r', locals())
                try:
                    o.addTranslationReference(canonical)
                except AlreadyTranslated as e:
                    _ERR('%(o)r is already translated!', locals())
                    _ERR(str(CantAddTranslationReference(o, canonical)), {})
                    if 0: pp({
                        'o': o, 'canonical': canonical,
                        'identical?': canonical is o,
                        })
                    set_trace()
                    nochmal = 0
                    if nochmal:
                        o.addTranslationReference(canonical)
                    else:
                        # https://www.python.org/dev/peps/pep-0344/#explicit-exception-chaining
                        # should work, shouldn't it?! (for me, it won't)
                        raise CantAddTranslationReference(o, canonical) # from e
                else:
                    changed = True
    # ----------------------- ] ... canonical, with language implied ]

    if changed or not set_language:
        return changed, notes

    if set_language:
        _NFO('%(o)r: Setting language to %(language)r', locals())
        try:
            o.setLanguage(language)
        except AttributeError as e:
            _ERR('handle_language(%(o)r): %(e)r', locals())
            pp(locals(), 'Nochmal, mit Debugger:',
            'Fehler tritt auf in LinguaPlone (.getTranslation --> isCanonical)')
            set_trace()
            o.setLanguage(language)
        else:
            changed = True
    return changed, notes  # ---------------------- ] ... handle_language ]


def handle_layout(o, kwdict, created, do_pop=True):  # --- [ h.layout ... [
    """
    handle the layout matters; return a 3-tuple (changed, notes, updates)
    """
    changed = False
    notes = []
    updates = {}
    _DBG, _NFO, _ERR = make_miniloggers(notes)

    layout, do_set, do_get = extract_layout_switch(kwdict, do_pop=do_pop)
    found_layout = None
    if do_set:
        if not layout:
            do_set = False
            _ERR("Won't set layout because of missing value (%(layout)r)",
                 locals())
        else:
            found_layout = o.getLayout()
            if found_layout == layout:
                _NFO('Checked layout=%(layout)r', locals())
            else:
                o.selectViewTemplate(templateId=layout)
                _NFO('set layout to %(layout)r (was %(found_layout)r)', locals())
                changed = True
    if do_get:
        if found_layout is None:
            found_layout = o.getLayout()
        if found_layout:
            _NFO('Found layout to be %(found_layout)r', locals())
            updates['layout'] = found_layout
    return (changed, notes, updates)  # ----------- ] ... handle_language ]


def handle_menu(o, kwdict, created, do_pop=True):  # -- [ handle_menu ... [
    """
    Handle the language and canonical keyword arguments for the given object
    `o`.

    o -- the object
    kwdict -- the kwargs dictionary of the calling function (will be consumed)
    created -- a boolean to tell whether the given object has just been
               created, affecting a few defaults

    Return a 2-Tuple (changed, notes);
    `notes` is in turn a list of ('INFO', `text`) tuples.
    """
    changed = False
    notes = []
    _DBG, _NFO, _ERR = make_miniloggers(notes)

    switch_menu = extract_menu_switch(kwdict, created, do_pop=do_pop)
    if switch_menu is not None:
        val = not switch_menu
        try:
            set_to = o.setExcludeFromNav
        except AttributeError:
            pt = getattr(o, 'portal_type', None)
            if pt == 'Plone Site':
                # this is expected:
                _NFO('%(o)r (%(pt)r) has no setExcludeFromNav attribute'
                     % locals())
            else:
                _ERR('%(o)r (%(pt)r) has no setExcludeFromNav attribute!'
                     % locals())
        else:
            set_to(val)
            _NFO('%(o)r: setExcludeFromNav(%(val)r)', locals())
            changed = True
    else:
        _DBG('switch_menu is %(switch_menu)r', locals())

    return changed, notes  # -------------------------- ] ... handle_menu ]


# --------------------------------- [ depends on visaplan.plone.subportals ... [
def might_set_subportal(kwdict):
    """
    The given keyword arguments indicate that a subportal *might* be set.

    I might not, even if this function returns True, because:

    - the subportal is already added
    - the `set_subportal` key is callable and might return False

    The kwdict dictionary is not consumed.

    >>> dic1 = {'only_other_keys': 42}
    >>> might_set_subportal(dic1)
    False
    >>> dic1
    {'only_other_keys': 42}
    >>> dic2 = {'subportal': 'some-subportal-id'}
    >>> might_set_subportal(dic2)
    True

    If `set_subportal` is callable, it *might* return True,
    but we don't try it (we don't have an object):
    >>> dic3 = {'subportal': 'other-subportal-id',
    ...         'set_subportal': lambda: False}
    >>> might_set_subportal(dic3)
    True

    However, if no subportal is available, it can't be set:
    >>> dic4 = {'set_subportal': True}
    >>> might_set_subportal(dic4)
    False
    """
    get = kwdict.get
    subportal = get('subportal')
    set_subportal = get('set_subportal')
    if set_subportal is None:
        return bool(subportal)
    elif (not set_subportal  # False
           or not subportal):
        return False
    else:
        return True


def handle_subportal(o, kwdict, created, do_pop=True):
    """
    Unitracc-specific functionality (visaplan.plone.subportals)
    """
    changed = False
    notes = []
    _DBG, _NFO, _ERR = make_miniloggers(notes)

    pop = (kwdict.pop if do_pop
           else kwdict.get)
    subportal_given = 'subportal' in kwdict
    set_subportal = pop('set_subportal', None)
    if callable(set_subportal):
        set_subportal, subportals = set_subportal(o)
    else:
        subportal = pop('subportal')
        if set_subportal is None:
            set_subportal = bool(subportal)
        elif not subportal and set_subportal:
            _NFO("Won't set_subportal %(subportal)r", locals())
            set_subportal = False
        if set_subportal:
            found_subportals = o.getSubPortals()
            found_set = set(found_subportals)
            if isinstance(subportal, six_string_types):
                wanted_set = set([subportal])
            else:
                wanted_set = set(subportal)
            _DBG('%(o)r: registered for subportals %(found_subportals)r', locals())

            added_set = wanted_set - found_set
            if added_set:
                subportals = list(found_subportals) + list(added_set)
            else:
                _NFO('%(o)r: checked subportal %(subportal)r', locals())
                set_subportal = False

    if set_subportal:
        o.setSubPortals(subportals)
        return True, notes
    return False, notes
# --------------------------------- ] ... depends on visaplan.plone.subportals ]


# ------------------------------------- [ depends on visaplan.plone.search ... [
def handle_united_search(siblings, opt, src_lang):
    """
    Unitracc-specific functionality (visaplan.plone.search):

    Take a given local search configuration from the primary object
    (siblings[src_lang]) and modify it to contain the paths of the siblings as
    well; this configuration is then applied to all siblings.

    Any changes are done to the @@settings in the pickle storage
    and don't affect the catalog brains.
    """
    if src_lang is not None:
        otherval = opt.get('src_lang' )
        if otherval is not None:
            assert src_lang == otherval
    else:
        src_lang = opt['src_lang']
    logger = opt['logger']
    me = 'handle_united_search'
    if src_lang is None:
        logger.warn('%(me)s: no source language for %(siblings)s', locals())
        return 0
    siblings = siblings.copy()  # a simple copy should be sufficient
    src_o = siblings.pop(src_lang)
    # get complete configuration:
    settings = opt['portal'].getBrowser('unitraccsettings')
    localsearch_all = settings.get('localsearch', {})

    src_uid = src_o.UID()
    src_search_raw = localsearch_all.get(src_uid)
    if not src_search_raw:
        logger.warn('%(me)s: no localsearch settings'
                    ' for %(src_o)r (%(src_uid)s)', locals())
        return 0
    src_search = normalize_localsearch(src_o, src_search_raw)
    logger.info('%(me)s: localsearch for UID %(src_uid)r: %(src_search)r', locals())
    pp(('src_o:', src_o), ('src_uid:', src_uid), ('search (stored):', src_search_raw), ('search (normalized):', src_search))
    search_path = src_search_raw.get('path')  # '.' oder SPEC oder so

    NOPATH = src_search.get('path') == 'NOPATH'
    if NOPATH:
        logger.info("%(me)s: localsearch for %(src_o)r"
                    " doesn't contain a path restriction"
                    ' (%(src_search_raw)s)', locals())
        added_here = frozenset()

    src_search_exp = expand_localsearch(src_o, src_search)
    src_paths = set(src_search_exp['path'])  # full path specs
    updated_settings = {}
    src_changes = False

    for la, other_o in siblings.items():
        other_uid = other_o.UID()
        logger.info('%(me)s: UID %(other_uid)r --> %(other_o)r', locals())
        if other_uid == src_uid:
            logger.info('%(me)s: UID equals source UID %(src_uid)r', locals())
            continue

        other_search_raw = localsearch_all.get(other_uid)
        if not other_search_raw:
            logger.info('%(me)s: No localsearch for UID %(other_uid)r so far', locals())
            if not NOPATH:
                tup = other_o.getPhysicalPath()
                other_path = '/'.join(tup)
                added_here = set([other_path])
        else:
            other_search = normalize_localsearch(other_o, other_search_raw)
            updated_settings[other_uid] = other_search
            other_search_exp = expand_localsearch(other_o, other_search)
            other_paths = set(other_search_exp['path'])
            logger.info('%(me)s: localsearch for UID %(other_uid)r: %(other_search)r', locals())
            if other_search['path'] == 'NOPATH':
                if NOPATH:
                    logger.info('%(me)s: localsearch for %(other_uid)r specifies NOPATH.', locals())
                else:
                    logger.warn('%(me)s: localsearch for %(other_uid)r specifies NOPATH.'
                                ' This will be changed!', locals())
            else:
                txt = ', '.join(sorted(other_paths))
                logger.info('%(me)s: localsearch for %(other_uid)r specifies %(txt)s', locals())
                if NOPATH:
                    logger.warn('%(me)s: path restriction for %(other_uid)r will be removed!', locals())
                else:
                    missing_here = src_paths - other_paths
                    if missing_here:
                        txt = ', '.join(sorted(missing_here))
                        # will be done below:
                        logger.info('%(me)s: localsearch for %(other_uid)r: adding %(txt)s', locals())

                    added_here = other_paths - src_paths
                    if added_here:
                        txt = ', '.join(sorted(missing_here))
                        logger.info('%(me)s: localsearch for %(other_uid)r: found %(txt)s', locals())
                        src_paths.update(added_here)
                        src_changes += 1

    if src_changes:
        src_search['path'] = sorted(src_paths)
        stored_dict = storable_localsearch(src_o, src_search)
        logger.info('%(me)s: Changing localsearch for %(src_uid)r to %(stored_dict)s', locals())

    for la, other_o in siblings.items():
        other_uid = other_o.UID()
        updated_settings.setdefault(other_uid, src_search)
        other_search = updated_settings[other_uid]
        paths_here = set(other_search['path'])
        paths_here.update(src_paths)
        other_search['path'] = sorted(paths_here)
        stored_dict = storable_localsearch(other_o, other_search)
        logger.info('%(me)s: Setting localsearch for %(other_uid)r to %(stored_dict)s', locals())
# ------------------------------------- ] ... depends on visaplan.plone.search ]



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et tw=79
"""
Mock-Klassen für Doctests
"""

# Python compatibility:
from __future__ import absolute_import

from six.moves import map

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"
VERSION = (0,
           4,  # MockBrowser, MockContext
           1,  # MockRequest(actual_url=...)
           )
__version__ = '.'.join(map(str, VERSION))
__all__ = ('MockPortal',
           'MockRequest',
           'MockObject',
           'MockBrain',
           'MockProfile',
           'MockLogger',
           'MockBrowser',
           )


class MockPortal:
    """
    Aus forms-Modul extrahiert (dort nicht mehr benötigt).
    """
    def absolute_url(self):
        return 'http://test.me'


class MockRequest(dict):
    def __init__(self, referer=None, **kwargs):
        self['HTTP_REFERER'] = referer
        self['ACTUAL_URL'] = kwargs.pop('actual_url', referer)
        self.form = dict(kwargs)


class MockObject(object):
    """
    Any persistent object in the ZODB object tree
    (which might be added to some catalog)

    Currently this doesn't support any special features:
    >>> o = MockObject()
    >>> o
    <an object>

    ... but it is considered seekable:
    >>> o.getHereAsBrain()
    <a brain>
    """
    def __repr__(self):
        return '<an object>'

    def getHereAsBrain(self):
        return MockBrain()


class MockBrain(dict):
    """
    Brain-Objekte haben folgende Methoden:
    - has_key bzw. __contains__

    Der Zugriff funktioniert aber nicht mit b.get(key),
    sondern nur mit getattr(b, key) ...

    MockBrains point to (currently: very basic) MockObjects:
    >>> MockBrain().getObject()
    <an object>
    """

    def __init__(self, **kwargs):
        dict.__init__(self, **kwargs)
        for (k, v) in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           ', '.join(['%s=%r' % (key, self[key])
                                      for key in sorted(self.keys())
                                      ]))

    def __repr__(self):
        """
        This is currently very basic
        (satisfying the needs of setup/_args.py):

        >>> MockBrain()
        <a brain>
        """
        return '<a brain>'

    def getObject(self):
        return MockObject()


class MockProfile:
    """
    Benutzerprofil;
    aus ..browser.registration.utils:
    """
    def UID(self):
        return 'abc123'


class MockLogger(list):
    """
    für Testzwecke; schreibt 2-Tupel in eine Liste und verhält sich ansonsten als solche.

    >>> logger = MockLogger()
    >>> logger.info('Eine Info (%(eins)s)', {'eins': 'zwei'})
    >>> logger.error('%d Fehler', 3)
    >>> list(logger)
    [('INFO', 'Eine Info (zwei)'), ('ERROR', '3 Fehler')]
    """
    def _cook(self, txt, *args):
        if args:
            if not args[1:] and isinstance(args[0], dict):
                return txt % args[0]
            return txt % args
        else:
            return txt

    def error(self, txt, *args):
        self.append(('ERROR', self._cook(txt, *args)))

    def info(self, txt, *args):
        self.append(('INFO', self._cook(txt, *args)))


class MockContext:
    # siehe auch MockProfile
    def UID(self):
        return 'MockUID_' * 4


class MockBrowser:
    """
    Für Doctests von Browsermethoden
    """
    context = MockContext()


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
attools - Archetype tools

Products.Archetypes should be installed to use this module;
otherwise, some functions will fail
"""

# Python compatibility:
from __future__ import absolute_import

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

__all__ = [
    'initialize_rich_text_fields',  # ./_at0.py
    'notifyedit',                   # ./_at1.py
    # ./_at2.py:
    'make_fields_inspector',
    'make_mimetype_fixer',
    'make_minilogger',
    'make_skip_function',
    # ./_at3.py:
    'getter_name',
    'getter_tuple',
    'instance_and_label',
    'setter_name',
    'setter_tuple',
    'get_first_text_as_html',       # ./_at4.py
    'generate_all_texts',           # ./_at5.py
    'get_all_texts',                # ./_at5.py
    'can_be_html',                  # ./_at6.py
    ]

# Local imports:
from ._at0 import initialize_rich_text_fields
from ._at1 import notifyedit
from ._at2 import (
    make_fields_inspector,
    make_mimetype_fixer,
    make_minilogger,
    make_skip_function,
    )
from ._at3 import (
    getter_name,
    getter_tuple,
    instance_and_label,
    setter_name,
    setter_tuple,
    )
from ._at4 import get_first_text_as_html
from ._at5 import generate_all_texts, get_all_texts
from ._at6 import can_be_html



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
(Mock-Version, nur für Testzwecke)

unitracc.tools.cfg: Zugriff auf "Produkt"konfiguration ...

... in buildout-Skript:

    zope-conf-additional +=
        <product-config name>
            debug_active 1
        </product-config>

(kopiert nach ../../../parts/instance/etc/zope.conf, führender Leerraum wird
entfernt)

ACHTUNG - Der Defaultwert für get_debug_active muß ein String sein; z. B.:

    from Globals import DevelopmentMode
    from visaplan.plone.tools.cfg import get_debug_active
    debug_active = get_debug_active('tree', str(DevelopmentMode))

"""

# Python compatibility:
from __future__ import absolute_import

__all__ = ['get_debug_active',
           'get_raw_config',
           ]

# visaplan:
from visaplan.tools.minifuncs import makeBool


class MockConfiguration(object):
    product_config = {}


def getConfiguration(*args, **kwargs):
    """
    Dummy: gib ein leeres dict zurück
    """
    return MockConfiguration()

def get_debug_active(product, default=None):
    """
    Lies den Konfigurationswert 'debug_active' für das übergebene Produkt aus.
    Das Ergebnis ist ein Wahrheitswert oder eine Ganzzahl.
    """
    dic = getConfiguration().product_config.get(product, {})
    val = dic.get('debug_active', '')
    return makeBool(val, default)


def get_raw_config(product, defaults={}):
    """
    Lies die Konfiguration für das übergebene Produkt aus
    und gib das "rohe" Dictionary zurück (ohne jegliche Wertekonversion).

    Schlüssel, die in der Konfiguration fehlen,
    werden ggf. dem defaults-dict entnommen.
    """
    dic = getConfiguration().product_config.get(product, {})
    if defaults:
        defaults.update(dic)
        return defaults
    else:
        return dic



# detect optional packages

# Python compatibility:
from __future__ import absolute_import

from importlib_metadata import PackageNotFoundError, version

try:
    version('beautifulsoup4')
except PackageNotFoundError:
    HAS_BEAUTIFULSOUP = False
else:
    HAS_BEAUTIFULSOUP = True

try:
    version('plone.app.blob')
except PackageNotFoundError:
    HAS_BLOB = False
else:
    HAS_BLOB = True

try:
    version('plone.dexterity')
except PackageNotFoundError:
    HAS_DEXTERITY = False
else:
    HAS_DEXTERITY = True

try:
    version('Products.Archetypes')
except PackageNotFoundError:
    HAS_ARCHETYPES = False
else:
    HAS_ARCHETYPES = True

try:
    version('visaplan.kitchen')
except PackageNotFoundError:
    HAS_KITCHEN = False
else:
    HAS_KITCHEN = True

try:
    version('visaplan.plone.infohubs')
except PackageNotFoundError:
    HAS_INFOHUBS = False
else:
    HAS_INFOHUBS = True

try:
    version('visaplan.plone.subportals')
except PackageNotFoundError:
    HAS_SUBPORTALS = False
else:
    HAS_SUBPORTALS = True

try:
    version('visaplan.plone.search')
except PackageNotFoundError:
    HAS_VPSEARCH = False
else:
    HAS_VPSEARCH = True

try:
    version('zope.i18n')
except PackageNotFoundError:
    HAS_ZOPE_I18N = False
else:
    HAS_ZOPE_I18N = True



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _tree
"""

# Python compatibility:
from __future__ import absolute_import

# Standard library:
from posixpath import normpath

# Zope:
from Products.CMFCore.utils import getToolByName

# Plone:
from plone.uuid.interfaces import IUUID

# Local imports:
from visaplan.plone.tools._have import HAS_SUBPORTALS
from visaplan.plone.tools.setup._args import (
    _extract_move_args,
    extract_layout_switch,
    extract_menu_switch,
    normalize_menu_switch,
    )
from visaplan.plone.tools.setup._misc import _traversable_path
from visaplan.plone.tools.setup._o_tools import (
    handle_language,
    handle_layout,
    handle_menu,
    handle_title,
    make_notes_logger,
    )

if HAS_SUBPORTALS:
    # Local imports:
    from visaplan.plone.tools.setup._o_tools import handle_subportal

# Local imports:
from .._idxnames import getExcludeFromNav
from visaplan.plone.tools.setup._reindex import make_reindexer

# Logging / Debugging:
import logging
from pdb import set_trace

__all__ = [
        'make_object_getter',
        ]


# see also _make_folder.py: make_subfolder_creator
def make_object_getter(context, **kwargs):
    """
    Return a function which finds an object ...

    - by 'id', if a 'parent' is given;
    - by 'uid' (if given);
    - by 'path' (relative to the portal object).

    Getting the object is the first part only; the function is used to make
    sure the object has certain properties as well!

    Further understood options are:
    - title - a title to be checked and/or set
    - language - a language to be checked and/or set
    - canonical - an object which was returned by some .getCanonical method
                  call. Usually requires a language value as well.
    - info - a (usually empty) dict; see `return_tuple` below.

    Options to the factory:
    - keys - the sequence of the keys 'id', 'uid' and 'path',
             specifying the order in which they are tried
    - verbose - log informations about problems
    - logger - a logger to use if verbose
    ... defaults for the function:
    - set_title - set the title (if given and not matching)
    - set_uid - set the UUID (if given and not matching)
    - set_language - set the language (if given and not matching)
    - set_canonical - link the canonical translation (if given)
    - reindex - True:  reindex in any case,
                False: ... under no circumstances,
                None:  ... if changes were made (default).
    - return_tuple - if True, return a 2-tuple (object, info);
                     by default, only the object (or None) is returned.

    Unless return_tuple=True is specified,
    the returned function will simply return the object or None;
    in this (standard) case, you can deliver the info dictionary yourself,
    which will be changed in-place, to get access to the
    detailed information, including an "updates" subdict.

    The "set_..." options mean, "set the object property", e.g. call
    setLanguage if a language key is given and mismatching;
    the "get_..." options do the opposite: they write the found value to the
    info dictionary, to be more precise: to the info['updates'] dictionary.
    Thus, to make use of the get_... results, an "info" dictionary must be
    provided.
    """
    pop = kwargs.pop
    keys = pop('keys', None) or ['path', 'id', 'uid']
    parent = pop('parent', None)
    portal = getToolByName(context, 'portal_url').getPortalObject()
    reference_catalog = getToolByName(context, 'reference_catalog')
    verbose = pop('verbose', 1)
    if 'logger' in kwargs:
        logger = pop('logger')
    elif verbose:
        logger = logging.getLogger('get_object')

    set_menu =  normalize_menu_switch(kwargs)
    reindexer = pop('reindexer', None)
    idxs =      pop('idxs', None)
    if idxs is None:
        idxs = []
    elif not idxs:  # e.g. False
        idxs = None
    mrx_kw = {  # make_reindexer keyword args
        'logger': logger,
        'context': context if parent is None
                   else parent,
        'idxs': idxs,
        }
    if idxs is None:  # defaults suppression;
        idxs = []     # won't be used! (see ignored_idxs below)
    reindex      = pop('reindex', reindexer is not None
                                  or None)
    set_title    = pop('set_title', True)
    set_uid      = pop('set_uid', False)
    get_uid      = pop('get_uid', None)
    set_language = pop('set_language', True)
    set_canonical = pop('set_canonical', None)
    if set_canonical is None:
        set_canonical = set_language
    set_subportal = pop('set_subportal', None)
    subportal     = pop('subportal', None)
    return_tuple  = pop('return_tuple', False)

    if set_menu is not None:
        idxs.append(getExcludeFromNav)  # changed to 'exclude_from_nav'
    if set_title:
        idxs.extend([
            'Title',
            'getTitleIndex',  # ??
            'sortable_title',
            'SearchableText',    # contains the title
            'getEfectiveIndex',  # contains lower-cased title after numeric time
            ])
    if set_uid:
        idxs.extend([
            'UID',
            ])
    if set_language:
        idxs.extend([
            'Language',
            ])
    if set_subportal:
        idxs.extend([
            'get_sub_portal',
            ])
    if reindex is None:
        reindex = (reindexer is not None
                   or bool(idxs)
                   )
    ignored_idxs = idxs is not mrx_kw['idxs']
    if (reindex
        and ignored_idxs
         or (idxs and reindexer is not None)
        ):
        logger.warn('Ignoring idxs value %(idxs)r', locals())

    if reindex and reindexer is None:
        reindexer = make_reindexer(**mrx_kw)

    def _err(msg, notes, logger=logger):
        notes.append(('ERROR', msg))
        if logger is not None:
            logger.error(msg)

    def _info(msg, notes, log=True, logger=logger):
        notes.append(('INFO', msg))
        if log and logger is not None:
            logger.info(msg)

    def get_object(
            id=None, uid=None, path=None,
            info=None,  # specify an empty dict to get information!
            parent=parent,
            reindex=reindex,
            reindexer=reindexer,
            set_title=set_title,
            set_uid=set_uid,
            get_uid=get_uid,
            set_language=set_language,
            set_canonical=set_canonical,
            set_subportal=set_subportal,
            subportal=subportal,
            return_tuple=return_tuple,
            **kwargs):
        """
        This function is designed to be called with keyword arguments only.

        In a Python-3-only release of the package, this will be enforced!
        """
        if info is None:
            info = {}
        info.update({
            'found':     False,
            'reindexed': False,
            'changes':   0,
            'notes':     [],
            'specs':     None,  # set to a string below
            'updates':   {},
            })
        notes = info['notes']
        # for notes from _o_tools.py:
        lognotes = make_notes_logger(logger, info['notes'])

        updates = info['updates']
        tried_keys = set()
        o = None
        found_by = None
        changes = 0
        specs = []
        for key in keys:
            if key in tried_keys:
                continue
            tried_keys.add(key)
            if key == 'id':
                if id is not None:
                    specs.append('id=%(id)r' % locals())
                    if parent is not None:
                        o = getattr(parent, id, None)
                        if o is None:
                            _err('%(parent)r.%(id)r not found!' % locals(),
                                 notes)
            elif key == 'uid':
                if uid is not None:
                    specs.append('uid=%(uid)r' % locals())
                    o = reference_catalog.lookupObject(uid)
                    if o is None:
                        _err('UID %(uid)r not found!' % locals(),
                             notes)
            elif key == 'path':
                if path is not None:
                    specs.append('path=%(path)r' % locals())
                    if normpath(path) in ('.', '/'):
                        o = portal
                        _info('path=%(path)r -> using %(portal)r!' % locals(),
                              notes)
                    else:
                        try:
                            # restrictedTraverse dislikes leading slashes, at least:
                            o = portal.restrictedTraverse(_traversable_path(path))
                        except KeyError:
                            o = None
                    if o is None:
                        _err('%(portal)r[%(path)r] not found!' % locals(),
                             notes)
            else:
                _err('Unknown key: %(key)r' % locals(), notes)

            if found_by is None and o is not None:
                found_by = key
                break

        info['specs'] = ', '.join(specs) or 'no non-empty specifications!'
        if o is None:
            return ((o, info) if return_tuple
                    else o)

        info['found'] = True
        info['found_by'] = found_by

        if found_by == 'uid':
            if get_uid >= 2:
                updates['uid'] = uid
        else:
            if get_uid is None:
                get_uid = True
            if get_uid or uid is not None:
                found_uid = IUUID(o, None)
                if uid is not None:
                    if found_uid == 'root':
                        if set_uid:
                            _err("Won't set UID of %(o)r to %(uid)r!"
                                 % locals(),
                                 notes)
                    elif found_uid != uid:
                        if set_uid:
                            o._setUID(uid)
                            _info('%(o)r: old UID %(found_uid)r --> new UID %(uid)r'
                                  % locals(),
                                  notes)
                            changes += 1
                            if get_uid:
                                found_uid = uid
                        else:
                            _info('%(o)r: UID %(found_uid)r mismatches %(uid)r'
                                  % locals(),
                                  notes)
                    else:
                        _info('%(o)r: checked UID (%(found_uid)r)' % locals(),
                              notes,
                              verbose > 1)
            if get_uid and found_uid != 'root':
                updates['uid'] = found_uid

        # ---------- [set_]title:
        kwargs.update(set_title=set_title)
        ch, notes = handle_title(o, kwargs, created=False)
        changes += ch
        for tup in notes:
            lognotes(tup)

        # ---------- [set_]language, [set_]canonical:
        kwargs.update(set_language=set_language, set_canonical=set_canonical)
        ch, notes = handle_language(o, kwargs, created=False)
        changes += ch
        for tup in notes:
            lognotes(tup)

        # ---------- [{set,get}_]layout:
        ch, notes, upd = handle_layout(o, kwargs, created=False)
        changes += ch
        for tup in notes:
            lognotes(tup)
        updates.update(upd)  # might contain a new 'layout' key

        # ---------- [switch_]menu:
        ch, notes = handle_menu(o, kwargs, created=False)
        changes += ch
        for tup in notes:
            lognotes(tup)

        if HAS_SUBPORTALS:
            # ---------- [set_]subportal:
            kwargs.update(subportal=subportal, set_subportal=set_subportal)
            ch, notes = handle_subportal(o, kwargs, created=False)
            changes += ch
            for tup in notes:
                lognotes(tup)

        info['changes'] = changes
        if reindex is None:
            if not changes:
                _info('%(o)r not changed and not reindexed' % locals(),
                      notes)
                return ((o, info) if return_tuple
                        else o)
            reindex = True
        if not reindex:
            if changes:
                _info('%(o)r has %(changes)d changes but reindexing suppressed'
                      % locals(),
                      notes)
            return ((o, info) if return_tuple
                    else o)
        if reindexer is None:
            o.reindexObject()
        else:
            reindexer(o=o)
        info['reindexed'] = True
        return ((o, info) if return_tuple
                else o)

    return get_object



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _workflow
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types

# Standard library:
from collections import defaultdict

# Zope:
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.WorkflowCore import WorkflowException

# visaplan:
from visaplan.tools.classes import DictOfSets

# Local imports:
from visaplan.plone.tools.setup._roles import set_local_roles
from visaplan.plone.tools.setup._watch import make_watcher_function

# Logging / Debugging:
import logging
from pdb import set_trace

# Exceptions:


__all__ = [
        'make_transition_applicator',
        ]


TRANSITIONS_MAP = {}
for from_state, to_state, tr in (
    ('restricted', 'visible',    'make_visible'),
    ('restricted', 'published',  'make_public'),
    ('visible',    'published',  'make_public'),
    ('visible',    'restricted', 'make_restricted_again'),
    ('published',  'visible',    'make_visible_again'),
    ('published',  'restricted', 'make_restricted_again'),
    # für Vortragsseiten, die in Demo-Kursen benötigt werden:
    ('inherit',    'visible',    'make_visible'),
    ('inherit',    'published',  'make_public'),
    # nichts zu tun, aber kein Fehler:
    ('restricted', 'restricted', None),
    ('visible',    'visible',    None),
    ('published',  'published',  None),
    ):
    TRANSITIONS_MAP[(from_state, to_state)] = tr
def make_transition_applicator(**kwargs):  # ---- [ m._t._a. ... [ -[[
    """
    Erzeuge eine Funktion, die den Workflow-Status eines als brain übergebenes
    Objekts ändert; dabei kann der Zielstatus aus der UID ermittelt werden
    (siehe uids_tuples).

    Argumente:
    - {default_,}target_state - der Zielstatus, wenn nicht aus uids_tuples zu
                                ermitteln
    - uids_tuples - eine Liste von (set, status, funktion)-Tupeln:
                    set - ein Set von UIDs, die den <status> (oder "besser")
                          erhalten sollen;
                    funktion - eine für die Objekte mit dem Zielstatus <status>
                               aufzurufende Funktion (z. B. zur rekursiven
                               Ermittlung der UIDs der verwendeten Objekte)
    - transitions_map - siehe TRANSITIONS_MAP
    - returns - was soll die Funktion zurückgeben?
      'changed' - True, wenn geändert wurde, sonst False (Vorgabewert)
      'target' - True, wenn der Zielstatus der gewünschte ist, sonst False
      'both' - ein 2-Tupel (changed, target)
      'error' - wie 'changed', aber mit Exception z. B. bei unbekannten
                Transitionen
    - doit_function - eine Funktion, die (nach den sonstigen Prüfungen)
                      entscheidet, ob die Transition (jetzt) angewendet werden
                      soll; sollte im anderen Fall eine Datenstruktur füllen,
                      um die (verzögerten) Transitionen später nachzuholen.
                      Signatur: f(brain, target_state)
    - set_inherit - soll die Akquisition von Zugriffsrechten vom Elternobjekt
                    umgeschaltet werden?  Mögliche Werte:
                    True - einschalten
                    False - ausschalten
                    'auto' - einschalten für Zielstatus 'inherit' oder
                             'published', sonst ausschalten
                    'autorestrict' - wie 'auto',
                                     aber nichts tun für 'published' (Vorgabe)
                    None - nichts tun
                    Die etwaige Umschaltung wird ungeachtet des Ergebnisses
                    einer <doit_function> ausgeführt, nicht aber wenn die
                    Transition lt. der internen <done_sets> (Schlüssel:
                    Zielstatus) schon durchgeführt wurde.
    - localroles_function - insbesondere wichtig für den Zielstatus
                    'restricted': eine Funktion, die bestimmt,
                    - welche "Prinzipale" (hier i.d.R.: Gruppen)
                    - welche lokalen Rollen
                    - zugewiesen (Vorgabe) oder entzogen bekommen sollen;
                    Signatur: f(brain, target_state).
                    Als Rückgabewert wird eine Liste von (id, [Rolle,...],
                    add)-Tupeln erwartet;
                    siehe --> make_simple_localroles_function
    - set_best_status - soll für das jeweilige Objekt der "beste" Status
                        gesetzt, d.h. ein etwaiger done-Vermerk für einen
                        nachgeordneten Status ignoriert werden?
                        (Siehe DictOfSets und die durch <uids_tuples>
                        vorgegebene Reihenfolge)
                        Vorgabe: True, wenn <uids_tuples> übergeben und nicht
                        leer, ansonsten False
    - force - soll eine Transition durchgeführt werden, obwohl sie für das
              übergebene Katalogobjekt (Kriterium: UID) oder einen "besseren"
              (=öffentlicheren) Zielstatus bereits durchgeführt wurde?
              (verwendet internes Dict <done_sets>)
    - regard_current - soll eine Transition als erfolgreich betrachtet werden
                       (insbesondere incl. der <done_sets>), wenn (ungeachtet
                       etwaiger Fehler) das Ergebnis stimmt?
    - shortcircuit - Transition von vornherein nicht versuchen, wenn der
                     Zielstatus schon vorliegt. Kann viel Zeit sparen,
                     insbesondere bei Medien mit Vorschaubildern!

    Debugging-Optionen:

    - watched_uid_and_status - siehe --> make_watcher_function;
                     set_trace, wenn apply_transition sich anschickt, das
                     Objekt mit der UID <uid> auf <status> zu setzen
    - tell_about_uids - ein Dict, {<uid>: <label>, ...},
                        oder eine Liste [(<uid>, <label>), ...];
                     wenn angegeben, wird zusätzlich eine Funktion
                     zurückgegeben, die über die Größen der internen Sets
                     informiert und darüber, ob die hier angegebenen UIDs
                     jeweils darin enthalten sind.
                 ACHTUNG:
                     Der Rückgabewert ist dann also keine einzelne Funktion,
                     sondern ein 2-Tupel (<apply_transition>, <tell...>)!
    """
    # ]]------ [ make_transition_applicator: Argumente ... [
    status_set = defaultdict(set)
    target_sets = DictOfSets()
    status_func = {}
    # Behandlung schon erledigter ...
    done_sets = DictOfSets()
    force = kwargs.pop('force', False)
    shortcircuit = kwargs.pop('shortcircuit', True)

    regard_current = kwargs.pop('regard_current', True)
    if 'default_target_state' in kwargs:
        default_target_state = kwargs.pop('default_target_state')
    else:
        default_target_state = kwargs.pop('target_state', None)
    uids_tuples = kwargs.pop('uids_tuples')
    for tup in uids_tuples:
        theset, target_state, func = tup
        if target_state in done_sets:
            raise ValueError('duplicate status %(target_state)r!' % locals())
        elif not isinstance(target_state, six_string_types):
            raise ValueError('Workflow status expected! (%target_state()r)'
                             % locals())
        target_sets.add_set(target_state)
        done_sets.add_set(target_state)
        if theset is None:
            if default_target_state is None:
                raise ValueError('No set for %(target_state)r,'
                                 ' and no default_target_state!'
                                 % locals())
            elif target_state != default_target_state:
                raise ValueError('No set for %(target_state)r which doesn\'t match'
                                 ' default_target_state %(default_target_state)r!'
                                 % locals())
            else:
                theset = set()
        elif not isinstance(theset, set):
            theset = set(theset)
        status_set[target_state] = theset
        target_sets[target_state].update(theset)
        status_func[target_state] = func
    set_best_status = kwargs.pop('set_best_status', bool(uids_tuples))
    del uids_tuples
    # neue UIDs den target_sets nur hinzufügen, wenn der Zielstatus
    # in den ursprünglich angegebenen enthalten ist:
    target_values = target_sets.ordered_keys()
    transitions_map = kwargs.pop('transitions_map', TRANSITIONS_MAP)
    returns = kwargs.pop('returns', 'changed')
    r_pool = 'changed target both error'.split()
    if returns not in r_pool:
        raise ValueError('invalid "returns" value (%(returns)r); '
                         'one of %(r_pool)s expected'
                         % locals())
    logger = kwargs.pop('logger', None)
    if logger is None:
        logger = logging.getLogger('apply_transition')

    verbosity = kwargs.pop('verbosity', 1)

    doit_function = kwargs.pop('doit_function', None)
    if doit_function is not None:
        doit = None
    else:
        doit = kwargs.pop('doit', True)
        if doit not in (False, True):
            raise ValueError('invalid "doit" value (%(doit)r)'
                             % locals())

    # ---------------------- [ Lokale Rollen ... [
    set_inherit = kwargs.pop('set_inherit', 'autorestrict')
    if set_inherit is None:
        pass
    elif set_inherit not in (False, True, 'auto', 'autorestrict'):
        raise ValueError('invalid "set_inherit" value (%(set_inherit)r)'
                         % locals())
    localroles_function = kwargs.pop('localroles_function', None)
    # ---------------------- ] ... Lokale Rollen ]

    # ---------- [ für Debugging (set_trace) ... [
    if 'watched_uid_and_status' in kwargs:
        tup = kwargs.pop('watched_uid_and_status')
        if tup:
            watched_case = make_watcher_function(tup,
                                                 logger=logger,
                                                 msg='WATCHED CASE: uid = %(key)r'
                                                 ' --> status = %(val)r!')
            logger.warn('DEBUGGING: set_trace, wenn ein beobachteter Fall '
                        'eintritt!')
        else:
            watched_case = gimme_False
    else:
        watched_case = gimme_False
    # ---------- ] ... für Debugging (set_trace) ]
    # -------- ] ... make_transition_applicator: Argumente ]

    tell_about_uids = kwargs.pop('tell_about_uids', False)
    if tell_about_uids:
        if isinstance(tell_about_uids, dict):
            tell_about_uids = list(tell_about_uids.items())

    # ------------ [ m._t._a.: generierte Infofunktion ... [
    stored_hits = {'prev': None,
                   'current': [],
                   }
    def summary(label=''):
        """
        Gib eine Zusammenfassung aus;
        lies hierzu folgende Variablen aus der Closure:

        done_sets -- die Sets mit den Erledigungen
        target_sets -- die Sets mit den gefundenen (oder explizit angegebenen)
                       UIDs
        stored_hits -- ...
        """
        info = []
        if label:
            label = label.strip()
        elif label is None:
            label = ''
        if label:
            label = label.join((' (', ')'))

        current_hits = stored_hits['current']
        has_hits = False
        found_targets = False
        for title, dos in [   # DictOfSets
            ('Zielstatus-Sets', target_sets),
            ('DONE-Sets',       done_sets),
            ]:
            if len(dos):
                info.append(title)
                for status in dos.ordered_keys():
                    theset = dos[status]
                    contains = []
                    for key, txt in tell_about_uids:
                        if key in theset:
                            current_hits.append((key, status, 'done'))
                            if key != txt:
                                contains.append('%(txt)s (%(key)s)' % locals())
                            else:
                                contains.append(key)
                    L = len(theset)
                    line = '%(L)7d %(status)r' % locals()
                    if contains:
                        has_hits = True
                        line += '; contains '+ ', '.join(contains)
                    info.append(line)
        if info:
            info.insert(0, '')
            info.append('')
        else:
            info.append('<empty!>')
        logger.info('apply_transition: SUMMARY%s {{%s}}',
                    label,
                    '\n  '.join(info))
        prev_hits = stored_hits['prev']
        if prev_hits is None:
            res = has_hits
        else:
            res = current_hits != prev_hits
        stored_hits['prev'] = current_hits
        stored_hits['current'] = []
        return res
    # ------------ ] ... m._t._a.: generierte Infofunktion ]

    # --------- [ m._t._a.: generierte Arbeitsfunktion ... [
    def apply_transition(brain, target_state=None,
                         default_target_state=default_target_state,
                         doit=doit,
                         logger=logger,
                         verbosity=verbosity,
                         force=force,
                         regard_current=regard_current,
                         set_inherit=set_inherit,
                         set_best_status=set_best_status,
                         localroles_function=localroles_function,
                         shortcircuit=shortcircuit):
        """
        Wende eine Workflow-Transition an und gib einen oder mehrere
        Wahrheitswerte zurück (--> erzeugende Funktion, <returns>)

        brain - das jeweilige Objekt wird als Katalogobjekt erwartet!
        target_state - wenn None, werden zunächst die Sets aus der erzeugenden
                       Funktion durchsucht (--> uid_tuples)
        default_target_state - verwendet, wenn <target_state> nach etwaiger
                       Suche in den <uid_tuples> immer noch None
        doit - wenn None, wird die <doit_function> (siehe erzeugende Funktion)
               verwendet, um zu entscheiden, ob die Transition direkt
               durchgeführt werden soll, oder andernfalls das Objekt für die
               spätere Verarbeitung vorzumerken
        logger - der zu verwendende Logger (Vorgabe aus erzeugender Funktion)
        verbosity - Ausführlichkeit, dto.
        force - <done_sets> ignorieren (siehe erzeugende Funktion)
        regard_current - wenn der aktuelle schon dem Zielstatus entspricht,
                         dies auch im Fehlerfall als erfolgreich behandeln;
                         weitgehend obsolet durch --> shortcircuit
        set_inherit - ein Wahrheitswert oder 'auto{,restrict}', oder None.
                      Siehe oben.
                      Wenn None, wird die etwaige Akquisition lokaler Rollen
                      nicht modifiziert.
        set_best_status - Vorgabewert siehe oben
        shortcircuit - wenn der aktuelle schon dem Zielstatus entspricht,
                       die Aktion als erfolgreich durchgeführt betrachten
                       (wenn auch ohne Änderungen)
        """
        # TODO: add_viewers_group, für restricted:
        # - auch, wenn Status schon 'restricted' ist (Reparatur)
        # - abhängig z. B. von portal_type; also z. B. durch Funktion zu regeln
        current_state = brain.review_state
        changed = False
        target_ok = True
        done = False
        o = None
        uid = brain.UID
        if target_state is None:
            target_state = target_sets.first_hit(uid, default_target_state)
            if target_state is None:
                target_state = default_target_state
                if watched_case(uid, target_state):
                    set_trace()
            if target_state is None:
                logger.error('Unknown target_state for UID %(uid)r', locals())
                if returns == 'error':
                    raise ValueError('Unknown target_state for UID %(uid)r'
                                     % locals())
                else:
                    changed, target_ok = None, None
                    done = True
        else:
            if watched_case(uid, target_state):
                set_trace()
            if target_state in target_values:
                # Zielstatus wurde übergeben; dem korrekten Set hinzufügen:
                target_sets[target_state].add(uid)

        # ----------------------- [ Zielstatus bekannt ... [
        if (not done   # der Regelfall!
                       # der Zielstatus wurde übergeben,
                       # konnte aus einem der Sets ermittelt werden,
                       # oder es gab zumindest einen Vorgabewert:
            and target_state is not None
            ):
            o = brain.getObject()
            # Es werden die Statuus geordnet überprüft;
            # was schon auf 'published' gesetzt wurde, braucht für 'published',
            # 'visible' und etwaige weitere nicht mehr berücksichtigt zu
            # werden.
            if set_best_status:
                done_state = done_sets.first_hit(uid, target_state)
            else:
                done_state = done_sets.first_hit(uid)
            if done_state is not None:
                logger.info('UID %(uid)r (--> %(target_state)r)'
                            ' found as done for %(done_state)r',
                            locals())
                if not force:
                    done = True  # set_inherit ist davon unabhängig

            # ---- [ nicht erledigt lt. done-Sets ... [
            if not done:
                func = status_func.get(target_state, None)
                if func is not None:
                    func(brain)  # z. B. rekursive Ermittlung der UIDs!
                if doit is None:
                    doit = doit_function(brain, target_state) or False
                if doit and shortcircuit:
                    if current_state == target_state:
                        changed, target_ok = False, True
                        done_sets.add(uid, target_state)
                        doit = False
                        pt = brain.portal_type
                        logger.info('%(uid)r %(pt)r (%(current_state)r): '
                                    'Keine Aktion erforderlich',
                                    locals())
                transition = None
                if doit:
                    # nun die Transition ermitteln:
                    try:
                        transition = TRANSITIONS_MAP[(current_state, target_state)]
                    except KeyError:
                        if regard_current and (current_state == target_state):
                            logger.info('%(o)r: keine Transition %(current_state)r --> %(target_state)r,'
                                        'aber der Zielstatus stimmt schon',
                                         locals())
                            changed, target_ok = None, True
                            done_sets.add(uid, target_state)
                        else:
                            logger.error('%(o)r: keine Transition %(current_state)r --> %(target_state)r bekannt!',
                                         locals())
                            if returns == 'error':
                                raise
                            else:
                                changed, target_ok = None, current_state == target_state
            # ---- ] ... nicht erledigt lt. done-Sets ]
            if not done and transition is None:
                if verbosity >= 3:
                    logger.info('Objekt %(uid)r bleibt %(current_state)r', locals())
                changed, target_ok = False, current_state == target_state
                done = True
            # --------- [ <transition> ist nun gesetzt ... [
            if not done:  # transition hat jetzt einen verwendbaren Wert
                # --------------- [ WF-Transition ... [
                if doit is None:
                    doit = doit_function(brain, target_state)
                if doit:
                    if o is None:
                        o = brain.getObject()
                    if verbosity >= 2 or brain.portal_type == 'Folder':
                        logger.info('%(uid)r %(o)r (%(current_state)r):'
                                    ' %(transition)s) ...',
                                    locals())
                    wft = getToolByName(o, 'portal_workflow')
                    try:
                        res = wft.doActionFor(o, transition)
                    except WorkflowException as e:
                        if regard_current and (current_state == target_state):
                            logger.info('%(uid)r, Transition fehlgeschlagen,'
                                        'aber der Zielstatus %(target_state)r stimmt schon (%(o)r)',
                                         locals())
                            target_ok = True
                            done_sets.add(uid, target_state)
                        else:
                            logger.error('%(uid)r %(o)r, Transition %(transition)r: %(e)r',
                                         locals())
                            target_ok = False
                    else:
                        done_sets.add(uid, target_state)
                        if verbosity >= 1:
                            logger.info('%(uid)r %(o)r, %(transition)s: OK', locals())
                        changed, target_ok = True, True
                else:
                    o = None
                    changed, target_ok = None, False
                # --------------- ] ... WF-Transition ]
            # --- [ Berechtigungs-Akquisition ... [
            if set_inherit is not None:
                if set_inherit == 'auto':
                    set_inherit = target_state in ('published', 'inherit')
                elif set_inherit == 'autorestrict':
                    if target_state in ('published',):
                        set_inherit = None
                    elif target_state == 'inherit':
                        set_inherit = True
                    else:
                        set_inherit = False
                if set_inherit is not None:
                    if o is None:
                        o = brain.getObject()
                    sharing = o.restrictedTraverse('@@sharing')
                    act_ = set_inherit and 'activate' or 'deactivate'
                    logger.info('%(uid)r %(o)r (-> %(target_state)r): '
                                '%(act_)s permission inheritance ...',
                                locals())
                    if sharing.update_inherit(set_inherit):
                        changed = True
                        logger.info('%(uid)r %(o)r (-> %(target_state)r): '
                                    'permission inheritance %(act_)sd',
                                    locals())
                    elif verbosity >= 1:
                        logger.info('%(uid)r %(o)r (-> %(target_state)r): '
                                    'permission inheritance not changed',
                                    locals())
            # --- ] ... Berechtigungs-Akquisition ]
            if localroles_function is not None:
                if set_local_roles(brain=brain, o=o,
                                   func=localroles_function,
                                   target_state=target_state,
                                   logger=logger):
                    changed = True
            # ---- ] ... <transition> ist nun gesetzt ]
        # ----------------------- ] ... Zielstatus bekannt ]

        if returns == 'target':
            return target_ok
        elif returns == 'both':
            return changed, target_ok
        else:  # 'changed' oder 'error'
            return changed
    # --------- ] ... m._t._a.: generierte Arbeitsfunktion ]

    if tell_about_uids:
        return apply_transition, summary
    return apply_transition
    # ----------------------------- ] ... make_transition_applicator ]





# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Switch objects on/off for search and navigation use
"""

__all__ = [
        'switch_menu_item',
        'show_item',
        'hide_item',
        ]


def switch_menu_item(o, on, logger, strict=True):
    """
    (De-) Aktiviert ein Objekt (i.d.R. einen Ordner) für das Menü

    o -- das Objekt
    on - der Wahrheitswert; die Negation wird verwendet als neuer Wert für excludeFromNav
         NEU in Version 1.4+: keine Änderung, wenn None!
    logger - der Logger
    strict - wenn True (Vorgabe), werden für `on` nur Wahrheitswerte akzeptiert
    """
    if on is None:
        logger.info('%(o)r.switch_menu_item: not changed (%(on)r)', locals())
        return
    elif strict and on not in (True, 1, False, 0):
        raise ValueError('Boolean value expected (%(on)r; for %(o)r)'
                         % locals())
    val = not on
    o.setExcludeFromNav(val)
    o.reindexObject()
    logger.info('%(o)r: setExcludeFromNav(%(val)r)', locals())


def show_item(o, logger, on=True):
    """
    Stellt ein Objekt für die Suche zur Verfügung.

    o -- das Objekt
    on - der Wahrheitswert; die Negation wird verwendet als neuer Wert für excludeFromSearch
    logger - der Logger
    """
    val = not on
    o.setExcludeFromSearch(val)
    o.reindexObject()
    logger.info('%(o)r: setExcludeFromSearch(%(val)r)', locals())


def hide_item(o, logger):
    """
    Verbirgt ein Objekt vor der Suche.

    o -- das Objekt
    logger - der Logger
    """
    show_item(o, logger, False)



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _rename
"""

# Python compatibility:
# Plone, sonstiges:
from __future__ import absolute_import

# Zope:
from Products.CMFCore.utils import getToolByName

__all__ = [
        # 'make_mover',  (not yet implemented)
        'make_renamer',
        # data:
        'ACCEPT_ANY',
        ]


def make_mover(context=None, **kwargs):
    """
    Erzeuge eine Funktion, die Seiten verschiebt und/oder UIDs zuweist.
    Benötigt werden hierzu
    - der Portalkatalog
    - das Portal und seine ID
    Die übliche Vorgehensweise:
    1. Wenn eine UID übergeben wurde, suche nach dieser UID.
       Wenn das Objekt über diese UID gefunden wurde, muß es einen der
       aufgeführten Pfade haben (den neuen oder, sofern angegeben, einen der
       alten); ansonsten tritt ein Fehler auf.
       Wenn die UID keinen Treffer ergab, suche an den angegebenen Stellen (neu
       und ggf. alt):
       - wenn nichts gefunden, erstelle ein neues Objekt an der *neuen* Stelle;
       - wenn an einer alten Stelle gefunden, verschiebe das Objekt;
       - weise dem neuen, vorhandenen oder verschobenen Objekt die UID zu.
    """


ACCEPT_ANY = (0, 'ACCEPT ANY')
def make_renamer(**kwargs):
    """
    Erzeuge eine Funktion, die ein per UID angegebenes Objekt umbenennt;
    Voraussetzung: der Elternordner bleibt derselbe (also keine Verschiebung)
    """
    if 'catalog' not in kwargs:
        context = kwargs.pop('context')
        catalog = getToolByName(context, 'portal_catalog')
    else:
        catalog = kwargs.pop('catalog')
    logger = kwargs.pop('logger')
    verbose = kwargs.pop('verbose', False)
    get_object = kwargs.pop('get_object', False)

    def rename_object(uid=None, oldid=None, newid=None,
                      verbose=verbose,
                      get_object=get_object,
                      **kwargs):
        """
        For historical reasons, the 1st three arguments are uid, oldid,
        newid; now they are all technically optional ...

        Thus, please specify all options by name:

          o -- the object to change.
          brain -- a catalog brain pointing to o.
                   If both are given, brain.getObject() must be identical
                   to o.
          uid -- if neither o nor brain is given, uid is required

        Changes and constraints:

          oldid -- constrain the ID of the object to be renamed;
                   None (default) or (more explicit) ACCEPT_ANY to skip
                   this check
          newid -- the new ID. If None, or if the same as the current id,
                   no renaming is performed

          newtitle -- a new value for the title
          oldtitles -- a list of possible old values for the title.

          The function expects to get at least a newid or newtitle;
          if neither is given, a TypeError is raised.

        Other options:

          verbose -- be a little bit more verbose, especially for skipped
                     operations
          get_object -- always compute (if not given) the object, even if
                        not necessary for internal processing
                        (because no changes are applicable).
                        If not specified, the return value might be None.
        """
        o = kwargs.pop('o', None)
        brain = kwargs.pop('brain', None)
        spec = None  # how was the object specified?
        if o is None:
            if brain is None:
                if uid is None:
                    raise ValueError('One of uid, brain or o(bject) '
                                     'is required!')
                spec = 'UID %(uid)r' % locals()
                query = {'UID': uid}
                brains = catalog(query)
                if not brains:
                    raise ValueError('UID %(UID)r not found!', query)
                elif brains[1:]:
                    number = len(brains)
                    raise ValueError('%(spec)s: Too many hits! (%(number)d)'
                                     % locals())
                brain = brains[0]
            if get_object:
                o = brain.getObject()
                if spec is None:
                    spec = repr(o)
            if spec is None:
                spec = 'brain(%s)' % (brain.getPath,)
        elif brain is not None:
            o2 = brain.getObject()
            if o2 is not o:
                raise ValueError('Given brain for %(o2)r doesn\'t match '
                                 'the given object %(o)r!'
                                 % locals())
            spec = repr(o)

        oldtitles  = kwargs.pop('oldtitles', None) or []
        newtitle = kwargs.pop('newtitle', None)

        if kwargs:
            logger.error('%(spec)s: ignored arguments %(kwargs)r!', locals())

        if not newid and not newtitle:
            raise TypeError('%(spec)s: No new id, no new title; nothing to do!'
                             % locals())

        changes = 0
        try:
            # we do have at least either o or a brain
            assert o is not None or brain is not None

            if newtitle:
                if brain is not None:
                    title = brain.Title
                else:
                    title = o.title()
                if title == newtitle:
                    if verbose:
                        if o is None: o = brain.getObject()
                        logger.info('%(spec)s: New title %(newtitle)r already set',
                                    locals())
                else:
                    if oldtitles == ACCEPT_ANY or title in oldtitles:
                        if o is None: o = brain.getObject()
                        logger.info('%(spec)s: setting title to %(newtitle)r '
                                    '(old: %(title)r)',
                                    locals())
                        o.title = newtitle
                        changes += 1
                    else:
                        logger.error('%(spec)s: unexpected title %(title)r!',
                                     locals())

            if not newid:
                if verbose:
                    logger.info('%(spec)s: No ID change', locals())
            else:
                if brain is not None:
                    current_id = brain.getId
                else:
                    current_id = o.getId()
                if current_id == newid:
                    if verbose:
                        logger.info('%(spec)s: renaming to %(newid)r already done.',
                                    locals())
                elif current_id == oldid or oldid in (None, ACCEPT_ANY):
                    if o is None: o = brain.getObject()
                    ## when renaming, we need to refresh any child nodes:
                    #if brain is not None:
                    #    oldpath = brain.getPath
                    #else:
                    #    oldpath = '/'.join(o.getPhysicalPath())
                    parent = o.aq_parent
                    logger.info('%(spec)s: renaming'
                                ' from %(current_id)r to %(newid)r.',
                                locals())
                    parent.manage_renameObject(oldid, newid)
                    o = getattr(parent, newid)
                    changes += 1
                    if verbose:
                        logger.info('renaming of %(current_id)r to %(newid)r done.', locals())
                else:
                    logger.error('%(spec)s: unexpected ID %(current_id)r, skipping.', locals())
        except Exception as e:
            logger.error('%(spec)s: renaming of %(current_id)r to %(newid)r failed!\n%(e)r', locals())
            raise
        else:
            if changes:
                assert o is not None, ('We have changes, so we need to '
                                       'have had the object!')
                logger.info('%(spec)s: reindexing ...', locals())
                o.reindexObject()
        return o

    return rename_object




# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Decorators for upgrade steps

Common functionality of the decorators provided by this module:

- They expect the decorated function to accept a context and a logger argument.
  The logger argument might be omitted depending on the usage, so it is
  complemented by the decorator for this cases;
  thus, the decorated function doesn't need any code to handle a missing logger option.

- They use the given or created logger to log the time it took to run the function.

- They catch the KeyboardInterrupt to allow the function to be aborted but not
  abort the foreground Zope process (which will receive a special StepAborted
  exception in this case);
  all other exceptions are logged (including the time the function ran up to
  this point) and re-raised.
"""

# Python compatibility:
from __future__ import absolute_import

# Standard library:
from functools import wraps
from time import time

# Local imports:
from visaplan.plone.tools.env import worker_name

# Logging / Debugging:
import logging

__all__ = [
        # decorators:
        'step',
        'importStep',
        # noch unfertig:
        # 'upgrade_step',
        'make_step_decorator',
        # exceptions:
        'StepAborted',
        ]


class StepAborted(Exception):
    """
    Vom Dekorator @step geworfen, wenn in einem Migrationsschritt eine
    KeyBoard-Exception ausgelöst wurde
    """


def step(func):
    """
    Dekorator für Migrationsschritte:
    - ergänzt ggf. fehlendes logger-Argument
    - stoppt die Zeit
    - (experimentell:)
      erlaubt das Weiterlaufen der Zope-Instanz auch nach manuellem Abbruch
      des Migrationsschritts
    """
    # da die def-Anweisung hier nicht ausgeführt wird, ist eine implizite
    # "Impfung" der verpackten Funktion mit zusätzlichen Variablen (wie z.B.
    # "site") leider nicht möglich; dafür würden zusätzliche Argumente
    # benötigt, und mithin eine Änderung der Signatur.

    @wraps(func)
    def wrapper(context, logger=None):
        funcname = func.__name__
        if logger is None:
            label = ':'.join([_f for _f in [
                                    'setup',
                                    worker_name(),
                                    funcname,
                                    ] if _f])
            logger = logging.getLogger(label)
        _started = time()
        try:
            res = func(context, logger)
        except Exception as e:
            delta = time() - _started
            if isinstance(e, KeyboardInterrupt):
                logger.error('%(funcname)s aborted after %(delta)5.2f seconds',
                             locals())
                logger.exception(e)
                raise StepAborted
            else:
                logger.error('%(funcname)s: %(e)r after %(delta)5.2f seconds',
                             locals())
                logger.exception(e)
                raise
        else:
            delta = time() - _started
            logger.info('%(res)r <-- %(funcname)s (%(delta)5.2f seconds)', locals())
        return res

    return wrapper


def importStep(marker_file, verbose=False):
    """
    Return a decorator for importStep functions.

    Options:

      marker_file -- relative path to a marker file (required)

             Note: This path is not (yet?) checked for existence,
                   but this might change in a future release!

      verbose -- enables some informational output (default: False)

    Import steps are usually excecuted in the context of their own package only
    and thus usually start with something like::

      >>> if context.readDataFile('my.astonishing.addon-marker.txt') is None:
      ...    return                   # doctest: +SKIP

    The file in question is commonly located in some local directory
    in the package sources ('profiles/default/') relative to the module.

    For convenience, we provide a decorator which allows you to write things
    like::

      >>> @importStep('profiles/default/my.astonishing.addon-marker.txt')
      ... def setup_various(context):
      ...     portal = context.getSite()
      ...     do_fancy_things(portal)  # doctest: +SKIP

    That way you see the path to that file; but as the readDataFile method
    expects the filename only, it is shortened.

    To prevent errors, the importStep won't accept anything callable by itself:

    >>> def some_function(context): pass
    >>> importStep(some_function)   # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    TypeError: This function doesn't decorate any function directly
               but expects the name of a marker file and
               *returns* the decorator!

    We *always* want our importSteps to check a marker file, right?!
    (If not, you wouldn't use our decorator -- if any at all!)
    >>> importStep()   # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    TypeError: importStep() takes at least 1 argument (0 given)

    Other wrong usage should be prevented as well:
    >>> importStep(None)   # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    ValueError: Please specify the name of a marker file!
    >>> importStep(0)   # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    AttributeError: 'int' object has no attribute 'split'

    """
    if marker_file is None:
        raise ValueError('Please specify the name of a marker file!')
    elif callable(marker_file):
        raise TypeError('This function doesn\'t decorate any function directly'
            ' but expects the name of a marker file'
            ' and *returns* the decorator!')

    fn = marker_file.split('/')[-1]

    def my_decorator(f):
        if verbose:
            print('*** decorating %r: check for marker file %r' % (f, fn))
        @wraps(f)
        def wrapper(context, *args, **kwargs):
            if verbose:
                print('*** Check context %r for file %s ...' % (context, fn))
            if context.readDataFile(fn) is None:
                return
            if verbose:
                print('*** Now executing %r ...' % f)
            try:
                return f(context, *args, **kwargs)
            except KeyboardInterrupt:
                raise StepAborted

        return wrapper

    return my_decorator


def upgrade_step(destination, **kwargs):
    """
    Gib einen Dekorator zur einmaligen Verwendung zurück.
    Es wird als erstes Argument die Zielversion (destination)
    angegeben; weitere werden üblicherweise in einem Dict gesammelt:

    >>> deco_kwargs = {'package': __package__, 'module': __name__}
    >>> @upgrade_step(1001, **deco_kwargs)  # doctest: +SKIP
    ... def create_interesting_folder(context, logger):
    ...     pass

    Der Dekorator sorgt für das (je nach Aufruf fehlende) logger-Argument;
    protokolliert wird:
    - die Zielversion (destination)
    - der Name der Funktion
    - jeweils wenn übergeben:
      - der Name des Python-Packages (package)
      - der Name des Moduls (module)
      - die Revisionsnummer (rev), normalerweise als String

    Die Logik ist wie folgt:
    - upgrade_step wird aufgerufen und verarbeitet die übergebenen Argumente,
      um eine Funktion mit einem Argument <func> zu erzeugen,
      die die übergebene Funktion dekoriert
    - die so generierte Funktion bekommt die letztlich zu dekorierende Funktion
      übergeben und gibt eine Wrapper-Funktion zurück, die folgendes tut:
      - sie erzeugt einen Logger, falls nicht übergeben,
        und ergänzt so das an die dekorierte Funktion übergebene logger-Argument;
      - sie protokolliert das Paket (sofern package übergeben; dringend
        empfohlen!) und die Zielversion
      - sie protokolliert den Namen der dekorierten Funktion
      - sie ruft die dekorierte Funktion auf und merkt sich den Zeitpunkt
      - sie protokolliert den Rückgabewert und die Brutto-Laufzeit der
        dekorierten Funktion
        - im Fehlerfall werden entsprechende Informationen protokolliert
      - schließlich nochmals ein Protokolleintrag für das Paket und die
        Zielversion
    """

    def make_wrapper():
        """
        package = package
        destination = destination
        rev = rev
        """
        rev = kwargs.pop('rev', 0)
        package = kwargs.pop('package')
        module = kwargs.pop('module', None)
        logger = kwargs.pop('logger', None)
        logger_name = kwargs.pop('logger_name', 'setup:%(funcname)s')

        @wraps(func)
        def wrapper(context, logger=logger):
            funcname = func.__name__
            package = package
            destination = destination
            rev = rev
            if logger is None:
                logger = logging.getLogger(logger_name % locals())
            logger.info('[ updating %(package)s to version %(destination)s ... [',
                        locals())
            '''
                        {'package': package,
                         'destination': destination,
                        })
            '''
            _started = time()
            if rev:
                logger.info('%(funcname)s@%(rev)s started', locals())
            else:
                logger.info('%(funcname)s started', locals())
            try:
                res = func(context, logger)
            except Exception as e:
                delta = time() - _started
                if isinstance(e, KeyboardInterrupt):
                    logger.error('%(funcname)s aborted after %(delta)5.2f seconds',
                                 locals())
                    logger.exception(e)
                    logger.error('] ... update of %(package)s to version'
                                 ' %(destination)s aborted ]',
                                 locals())
                    raise StepAborted
                else:
                    logger.error('%(funcname)s: %(e)r after %(delta)5.2f seconds',
                                 locals())
                    logger.exception(e)
                    logger.error('] ... update of %(package)s to version'
                                 ' %(destination)s failed ]',
                                 locals())
                    raise
            else:
                delta = time() - _started
                logger.info('%(funcname)s completed (%(delta)5.2f seconds)', locals())
                logger.info('] ... %(package)s updated to version %(destination)s ]',
                            locals())
            return res

        return wrapper
    return make_wrapper()


def make_step_decorator(**kwargs):
    """
    Erzeuge einen Dekorator wie vorstehende Funktion --> step,
    aber ergänze bei der Protokollierung des Aufrufs, gemäß benannten
    Argumenten:
    - rev --> die svn-Revision des aufrufenden Moduls;
      z. B., wenn das svn:keyword "Revision" aktiv ist:
      >>> MODULE_REVISION = '$Revision: 31126 $'[8+3:-2]
    """
    rev = kwargs.pop('rev', 0)
    mask = 'setup:%(funcname)s'
    if rev:
        int(rev)
        mask += '@' + rev

    def make_wrapper(func, rev=rev):
        @wraps(func)
        def wrapper(context, logger=None):
            funcname = func.__name__
            rev = rev
            if logger is None:
                logger = logging.getLogger('setup:'+funcname)
            _started = time()
            if rev:
                logger.info('%(funcname)s@%(rev)s started', locals())
            try:
                res = func(context, logger)
            except Exception as e:
                delta = time() - _started
                if isinstance(e, KeyboardInterrupt):
                    logger.error('%(funcname)s aborted after %(delta)5.2f seconds',
                                 locals())
                    logger.exception(e)
                    raise StepAborted
                else:
                    logger.error('%(funcname)s: %(e)r after %(delta)5.2f seconds',
                                 locals())
                    logger.exception(e)
                    raise
            else:
                delta = time() - _started
                logger.info('%(funcname)s completed (%(delta)5.2f seconds)', locals())
            return res

        return wrapper
    return make_wrapper

if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()
